from nio.project import Project


class ProjectSerializer(object):

    """ A project serializer is a class that is intended to be extended to
    allow for the serialization and deserialization of n.io projects.

    Subclasses should implement a serialize and deserialize method to convert
    to and from nio.project.Project instances. It is generally assumed that
    passing a project through serialize and then deserialize will return
    the same project.
    """

    def serialize(self, project, service_persistence=False):
        """ Take a project and serialize it to some state.

        This method is intended to be implemented in subclassed Serializers.
        It will be passed an instance of a nio.project.Project class and
        it should convert that instance to its serialized version. For example,
        a file based serializer would take a Project instance and create the
        proper files on the file system that represent the project.

        Args:
            project (Project): An instance of a n.io project
            service_persistence (bool): Specifies if service persistence data 
                is serialized

        Returns:
            None
        """
        raise NotImplementedError

    def deserialize(self, service_persistence=False):
        """ Create a Project instance from this serializer.

        This method should be implemented in subclassed Serializers. It should
        return an instance of a nio.project.Project class that is
        representative of the project in the serialized state. For example,
        a file based project serializer would read the file system and parse
        the data into the Project class.

        Args:
            project (Project): An instance of a n.io project
            service_persistence (bool): Specifies if service persistence data
                is de-serialized

        Returns:
            project (Project): An instance of a n.io project
        """
        raise NotImplementedError

    def validate_project(self, project):
        """ Make sure a project is a valid n.io project.

        The default implementation of this method just makes sure that the
        project is of the right type. Subclasses could perform additional
        validation if desired.

        Args:
            project: A n.io project

        Returns:
            None: This method will raise an exception if the project is
                invalid. No return should be considered a success.

        Raises:
            TypeError: If the project is not a nio.project.Project
        """
        if not isinstance(project, Project):
            raise TypeError("Project is not a valid Project instance")



import logging
import sys
from enum import Enum

from nio.util.nio_time import get_nio_time


class NIOAdapter(logging.LoggerAdapter):

    """ The base logger adapter class for NIO.  """

    detail_errors = True

    def __init__(self, logger, extra={}):
        super().__init__(logger, extra)

    def setLevel(self, level):
        """ Set the specified level on the underlying logger.  """
        log_level = level.value if isinstance(level, Enum) else level
        super().setLevel(log_level)

    def exception(self, msg, *args, **kwargs):
        """ Log an exception inside of a handler.

        Delegate an exception call to the underlying logger and log exception
        or error depending of running flag "detail_errors"
        """
        if self.__class__.detail_errors:
            super().exception(msg, *args, **kwargs)
        else:
            _, exc_value, _ = sys.exc_info()
            msg = "{0}, (Exception details: {1})".format(
                msg, self.exc_info(exc_value))
            super().error(msg, *args, **kwargs)

    @classmethod
    def exc_info(cls, ex):
        """ Returns string information for an exception.  """
        try:
            if isinstance(ex, Exception) and hasattr(ex, 'message'):
                ex_str = "{0}{1}{2}".format(type(ex).__name__,
                                            " - " if ex.message else "",
                                            ex.message)
            else:
                ex_str = str(ex)
                ex_str = "{0}{1}{2}".format(
                    type(ex).__name__, " - " if ex_str else "", ex_str)
        except:
            ex_str = repr(ex)
        return ex_str

    def process(self, msg, kwargs):
        """ Add some additional context to our log record

        Args:
            msg (str): The message to be logged.

        Kwargs:
            see LoggerAdapter.process

        Returns:
            The (possibly modified) versions of the arguments passed in.
        """

        self.extra["niotime"] = get_nio_time()
        self.extra["context"] = self.logger.name
        return super().process(msg, kwargs)

    @classmethod
    def name_to_level(cls, name):
        """ Converts from level name to level identifier

        Args:
            name (str): Log name

        Returns:
            log level identifier
        """
        return getattr(logging, name.upper(), None)



from copy import deepcopy
from nio.signal.base import Signal
from nio.properties import StringProperty, BoolProperty, \
    ObjectProperty, PropertyHolder


class EnrichProperties(PropertyHolder):
    enrich_field = StringProperty(
        title="Results Field", default="", visible=False)
    exclude_existing = BoolProperty(title="Exclude Existing?", default=True)


class EnrichSignals(object):

    enrich = ObjectProperty(EnrichProperties, title='Signal Enrichment',
                            default=EnrichProperties(), order=100)

    def get_output_signal(self, signal_data, incoming_signal, copy=True):
        """ Get an output signal based on the block configuration.

        This method will return a single Signal that consists of the fields
        according to the block's configuration. The signal_data may be the
        entire contents of the signal, it may be added as a field, or it may
        be added to the contents of the signal.

        Args:
            signal_data (dict): Dictionary of data representing the "results"
                to append/include with the signal
            incoming_signal (Signal): The signal object that powered this
                particular operation
            copy (bool): Whether or not to perform a deep copy on the
                incoming signal before operating on it

        Returns:
            sig (Signal): A single Signal formatted according to the block
        """
        if self.enrich(incoming_signal).exclude_existing(incoming_signal):
            # This is the easy case, we don't want any of the previous data
            return Signal(signal_data)

        if copy:
            new_sig = deepcopy(incoming_signal)
        else:
            new_sig = incoming_signal

        if self.enrich().enrich_field():
            # Just set the signal data to a result attribute
            setattr(new_sig, self.enrich().enrich_field(), signal_data)
            return new_sig

        # If we're here, we need to merge the signal data onto the incoming
        # signal. Data from the results should take precedence over data on
        # the incoming signal.
        new_sig.from_dict(signal_data)
        return new_sig

    def notify_output_signals(self, signals_data, incoming_signal,
                              copy=True, output_id=None):
        """ A helper method to notify signals after properly enriching them.

        This method will call self.notify_signals on the block's behalf. It
        will first enrich the incoming signal if the block is configured to
        do so.

        Calling this method is roughly equivalent to calling get_output_signal
        and then notifying the result.

        Args:
            signals_data (dict/list): Data about the output signals. If this
                is a list, it will be assumed to be a list of dictionaries
                whose corresponding Signals will be notified in a batch.
                Otherwise, it should be a single dict containing the signal
                data to notify.
            incoming_signal (Signal): The signal object that triggered this
                operation.
            copy (bool): Whether or not to perform a deep copy on the
                incoming signal before operating on it.
            output_id (str): The output ID to notify signals to
        """
        if not isinstance(signals_data, list):
            signals_data = [signals_data]

        self.notify_signals([
            self.get_output_signal(signal_data, incoming_signal, copy=copy)
            for signal_data in signals_data], output_id)



from nio.types.base import Type


class StringType(Type):

    @staticmethod
    def serialize(value, **kwargs):
        """ Convert a value to a JSON serializable value """
        return value

    @staticmethod
    def deserialize(value, **kwargs):
        """ Convert value to string """
        try:
            return str(value)
        except:
            raise TypeError("Unable to cast value to a string")



import pickle
from nio.project.serializers.serializer import ProjectSerializer


class BinarySerializer(ProjectSerializer):

    """ A simple serializer that keeps a binary representation of a project
    in its memory. These bytes can be saved to disk, sent over the network, or
    really anything else you can do with bytes. The bytes are created using
    Python pickle so deserializing will require a compatible nio framework
    installed in the deserializing environment.

    To serialize a project call:
        >>> serializer = BinarySerializer()
        >>> serializer.serialize(project)
        >>> project_bytes = serializer.binary_data

    To deserialize bytes:
        >>> serializer = BinarySerializer(project_bytes)
        >>> project = BinarySerializer.deserialize()
    """

    def __init__(self, binary_data=None):
        # The bytes representing the project after serialization or before
        # deserialization
        self.binary_data = binary_data

    def serialize(self, project):
        """ Take a project and turn it into bytes """
        # Make sure the project is a valid one
        self.validate_project(project)

        self.binary_data = pickle.dumps(project)

    def deserialize(self):
        return pickle.loads(self.binary_data)



"""

   Service Context: The data that a service will be configured with

"""


class ServiceContext(object):

    def __init__(self, properties, blocks=None,
                 block_router_type=None, router_settings=None,
                 mgmt_signal_handler=None,
                 blocks_async_configure=True,
                 blocks_async_start=False,
                 blocks_async_stop=True,
                 instance_id=None):
        """ Initializes information needed for a Service

        Arguments:
            properties (dict): service metadata
            blocks (list): list of blocks, each with the format:
                {"type": block class,
                 "properties": block metadata}
            block_router_type: block router class to use
            router_settings (dict): router settings, , these can include
                "clone_signals" and/or any other settings depending on router
                being used
            mgmt_signal_handler (method): method to use to publish
                management signals, receives signal as only parameter
            blocks_async_configure: If True, blocks configure asynchronously
            blocks_async_start: If True, blocks start asynchronously
            blocks_async_stop: If True, blocks stop asynchronously
            instance_id: Instance this service belongs to
        """
        self.properties = properties
        self.blocks = blocks if blocks is not None else {}
        self.block_router_type = block_router_type
        self.router_settings = router_settings or {}
        self.mgmt_signal_handler = mgmt_signal_handler
        self.blocks_async_configure = blocks_async_configure
        self.blocks_async_start = blocks_async_start
        self.blocks_async_stop = blocks_async_stop
        self.instance_id = instance_id



from nio.modules.module import Module
from nio.modules.security.authenticator import Authenticator
from nio.modules.security.authorizer import Authorizer


class SecurityModule(Module):

    def proxy_authenticator_class(self, authn_class):
        Authenticator.proxy(authn_class)

    def proxy_authorizer_class(self, authz_class):
        Authorizer.proxy(authz_class)

    def finalize(self):
        Authenticator.unproxy()
        Authorizer.unproxy()
        super().finalize()

    def get_module_order(self):
        return 30

    def get_module_type(self):
        return "security"



from time import sleep
from nio.block.mixins.retry.strategy import BackoffStrategy


class LinearBackoff(BackoffStrategy):

    def wait_for_retry(self):
        """Sleep a certain number of seconds before we do the retry.

        We will wait the current retry number times the multiplier,
        so first retry will be 1*mult, second will be 2*mult, etc.

        The exception is when our current retry is greater than our max
        number of retries, then we want to use the max number of retries.
        This can happen if the strategy is configured to run indefinitely.
        """
        sleep(min(self.retry_num, self.max_retry) * self.multiplier)



from nio.block.mixins.collector.collector import Collector
from nio.block.mixins.enrich.enrich_signals import EnrichSignals
from nio.block.mixins.group_by.group_by import GroupBy
from nio.block.mixins.limit_lock.limit_lock import LimitLock
from nio.block.mixins.persistence.persistence import Persistence
from nio.block.mixins.retry.retry import Retry



from nio.router.base import BlockRouter
from concurrent.futures.thread import ThreadPoolExecutor


class ThreadedPoolExecutorRouter(BlockRouter):
    """ A router that delivers signals asynchronously in a thread pool

    This router caps the number of threads used for block's signal receiving.
    """

    def __init__(self):
        """ Create a new threaded pool executor router """
        super().__init__()
        self._executor = None

    def configure(self, context):
        """ Configures router

        Instantiates actual pool executor
        """
        super().configure(context)

        max_workers = context.settings.get("max_workers", 50)
        self._executor = ThreadPoolExecutor(max_workers=max_workers)

    def deliver_signals(self, block_receiver, signals):
        """ Deliver the signals inside of the thread executor """
        self._executor.submit(
            self.notify_signals_to_block, block_receiver, signals)



class Project(object):

    """ A class that contains all information needed for a n.io project.

    This class allows for us to create an in-memory representation of a project.
    That instance can be validated and/or serialized into different formats
    that will ultimately be read by a n.io binary.

    For the most part, a n.io project consists of a collection of configured
    blocks, services, and other configuration options. A special configuration
    option is the block types that are available to the project. Currently,
    this representation only allows for block types that are configured in the
    environment section of a configuration file.
    """

    def __init__(self):
        super().__init__()
        # A project has configured blocks. It is stored as a dictionary
        # where the key is the block id and the value is a Block instance
        self.blocks = dict()

        # A project has configured services. It is stored as a dictionary
        # where the key is the service name and the value is a Service instance
        self.services = dict()

        # A project has some configuration. The top-level dictionary keys are
        # the configuration sections, the values are configuration objects
        self.configuration = dict()

        # A project might contain persistence data besides blocks and services
        # at the core level, for example, a given core component might save
        # data using Persistence and would like such data to be part of
        # the project
        self.core_persistence = dict()

        # Data saved for a service using Persistence
        self.service_persistence = dict()


from nio.modules.proxy import ModuleProxy


class Settings(ModuleProxy):

    """ Interface class to store settings """

    @classmethod
    def get(cls, section=None, option=None, *args, **kwargs):
        """Accessor for settings.

        Args:
            section (str): section to get information from
            option (str): option to get information from

        if no section is specified, all sections are returned
        if no option is specified, all options within section are returned

        Keyword Args:
            fallback: default to use when no information can be found

        Returns:
            settings: The requested settings or
                None if no possible value was found
        """
        raise NotImplementedError()

    @classmethod
    def getint(cls, section, option, **kwargs):
        # a convenience method converting section option value to an integer
        raise NotImplementedError()

    @classmethod
    def getfloat(cls, section, option, **kwargs):
        # a convenience method converting section option value to a float
        raise NotImplementedError()

    @classmethod
    def getboolean(cls, section, option, **kwargs):
        # a convenience method converting section option value to a boolean
        raise NotImplementedError()

    @classmethod
    def getdict(cls, section, option, **kwargs):
        # a convenience method reading the section option value interpreting
        # it as a python dictionary
        raise NotImplementedError()

    @classmethod
    def set(cls, section, option=None, value=None):
        """ Sets the value for a given section and option

        if the section does not exists, it creates it
        if option is not specified, value is assigned to the section and it
            must be a dictionary type

        Args:
            section (str): section to set information to
            option (str): option to set information to
            value: value to assign

        Raises:
            TypeError if value is invalid
        """
        raise NotImplementedError()

    @classmethod
    def clear(cls):
        """ Clears all settings """
        raise NotImplementedError()



from nio.properties.base import BaseProperty
from nio.properties.exceptions import NoClassVersion, NoInstanceVersion, \
    OlderThanMinVersion
from nio.util.versioning.check import compare_versions, \
    VersionCheckResult, InvalidVersionFormat, is_version_valid, \
    get_major_version


class PropertyHolder(object):

    """ Support nio.properties.BaseProperty as class attributes.

    Functionality to group all properties of a given class.
    Provides methods to serialize and deserialize a given instance, and
    to obtain the description at the class level.

    Serializing an instance is useful when saving instance settings
    De-serializing an instance is useful when loading instance settings
    Obtaining a description is useful when exposing the class properties

    """

    def to_dict(self):
        """ Return a dictionary representation of itself.

        Args:
            None

        Returns:
            Instance serialization as a dictionary

        """
        class_properties = self.__class__.get_class_properties()
        return {property_name: prop.serialize(self)
                for (property_name, prop) in class_properties.items()}

    def validate(self, ignore_none=False):
        """ Call and deserialize each input property to determine validity.

        Args:
            ignore_none (bool): if True, properties that have a value equal to
                None are ignored (useful when wanting to validate errors on
                properties for which a value other than None has been assigned)

        Raises:
            AllowNoneViolation: Property value does not allow none, available
                when ignore_none is False
            TypeError: Property value is invalid

        """
        class_properties = self.__class__.get_class_properties()
        for (property_name, prop) in class_properties.items():
            # Deserialize the raw value of the PropertyValue
            value = getattr(self, property_name).value
            if value is None and ignore_none:
                continue
            # Deserialize to check for AllowNoneViolation and TypeError
            prop.deserialize(value)

    @classmethod
    def validate_dict(cls, properties):
        """ Call and deserialize each input property to determine validity.

        Validates the given property dictionary by successively
        de-serializing each property, returning the resulting (validated)
        property dictionary. If no exceptions are thrown here, all
        properties are valid.

        Args:
            properties (dict): values to validate

        Returns:
            properties (dict): validated and serialized

        Raises:
            AllowNoneViolation: Property value does not allow none
            TypeError: Property value is invalid

        """
        class_properties = cls.get_class_properties()
        for (property_name, prop) in class_properties.items():
            if property_name in properties:
                value = properties[property_name]
                # Deserialize to check for AllowNoneViolation and TypeError
                prop.deserialize(value)
                # Return the serialized version of the input dictionary
                serialized_value = prop.type.serialize(value, **prop.kwargs)
                properties[property_name] = serialized_value
        return properties

    def from_dict(self, properties, logger=None):
        """ Load properties from the specified dict into the instance.

        Note: Existing values for properties that are not included in
        the properties dict would remain.

        Args:
            properties (dict): values to assign to this instance

        Returns:
            None

        """
        # perform minimum validation
        if properties is None:
            raise TypeError()

        # Retrieve the list of all class properties
        class_properties = self.__class__.get_class_properties()
        self._process_and_log_version(class_properties, properties, logger)
        for (property_name, prop) in class_properties.items():
            if property_name in properties:
                setattr(self, property_name, properties[property_name])

    @classmethod
    def get_description(cls):
        """ Provide the instance properties.

        Args:
            None

        Returns:
            Instance description as a dictionary of properties

        """
        class_properties = cls.get_class_properties()
        descriptions = {property_name: prop.description
                        for (property_name, prop) in class_properties.items()}
        if hasattr(cls, "__version__") and "version" not in descriptions:
            descriptions["version"] = cls.__version__

        return descriptions

    @classmethod
    def get_defaults(cls):
        """ Determine the instance properties and their default values.

        Args:
            None

        Returns:
            (dict): The default values for all properties, indexed
                by name.
        """
        properties = cls.get_class_properties()
        return {prop_name: prop.default
                for prop_name, prop in properties.items()}

    @classmethod
    def get_serializable_defaults(cls):
        """ Determine the instance properties and their serializable defaults.

        Args:
            None

        Returns:
            (dict): The serializable default values for all properties, indexed
                by name.
        """
        properties = cls.get_class_properties()
        return {prop_name: prop.serialize(instance=cls)
                for prop_name, prop in properties.items()}

    @classmethod
    def get_class_properties(cls):
        """ Determine the metadata properties on this class.

        Determine the metadata properties on this class by means of reflection.
        This is useful in serialization and deserialization.

        Args:
            None

        Returns:
            class_properties (dict): The discovered properties, indexed
                by name.

        """
        class_attribute = "{0}_properties".format(cls.__name__)
        if not hasattr(cls, class_attribute):
            # find out properties
            properties = dict()
            import inspect
            classes = inspect.getmro(cls)
            for _class in classes:
                for (prop_name, prop) in _class.__dict__.items():
                    if (isinstance(prop, BaseProperty) and
                            prop_name not in properties):
                        properties[prop_name] = prop

            # cache properties
            setattr(cls, class_attribute, properties)
        return getattr(cls, class_attribute)

    def _process_and_log_version(self, class_properties, properties, logger):
        # get either 'id' or 'name' property of this holder
        id = properties.get("id", properties.get("name", ""))
        try:
            self._handle_versions(class_properties, properties)
        except OlderThanMinVersion as e:
            if logger:
                logger.warning('Instance {0} version: {1} is older than'
                               ' minimum: {2}'.format
                               (id, e.instance_version, e.min_version))
        except (NoClassVersion, NoInstanceVersion):
            # pass on classes with no version info.
            # pass on instances with no version info in their config.
            pass

    def _handle_versions(self, class_properties, instance_properties):
        """ Raise version exceptions based on instance config.

        Determine version relation of an instance with respect to class version
        definition.

        Assumes that both, class and instance refer to their version through
        a version property

        Args:
            class_properties: class properties
            instance_properties: instance properties as retrieved likely from
                configuration file.

        Raises:
            NoClassVersion
            NoInstanceVersion
            InvalidVersionFormat
            OlderThanMinVersion
        """

        try:
            class_version = class_properties.get("version").default
        except:
            raise NoClassVersion()

        # need to check against possible version stored in properties
        if "version" not in instance_properties:
            raise NoInstanceVersion(class_version)

        # compare versions and determine compatibility
        instance_version = instance_properties["version"]

        if not is_version_valid(instance_version):
            raise InvalidVersionFormat()

        comparison_result = compare_versions(instance_version,
                                             class_version)
        if comparison_result == VersionCheckResult.equal \
                or comparison_result == VersionCheckResult.newer:
            # instance has a newer version, it is ok
            pass
        else:
            # got an older version in the block, check if it passes
            # min version
            if "min_version" not in class_properties["version"].kwargs:
                # min_version by default is a major version, which is built
                # starting from the major digit, and adding zero's to it.
                min_version = get_major_version(class_version)
            else:
                min_version = \
                    class_properties["version"].kwargs["min_version"]

            comparison_result = compare_versions(instance_version,
                                                 min_version)
            if comparison_result == VersionCheckResult.equal \
                    or comparison_result == VersionCheckResult.newer:
                # ok, got a version not older than minimum
                pass
            else:
                raise OlderThanMinVersion(instance_version,
                                          min_version)



from nio.util.scheduler.scheduler import SchedulerRunner


class JumpAheadSchedulerRunner(SchedulerRunner):
    """ Overrides SchedulerHelper allowing simulation of forward jumps in time
    """
    def __init__(self):
        super().__init__()
        # allows to simulate jumps in time during testing
        self.offset = 0

    def jump_ahead(self, seconds):
        """ Simulate a jump forward in time

        This will update the scheduler's offset a certain number of seconds
        in order to simulate time passing. Any scheduled jobs that should have
        fired during the time that passed will fire after scheduler thread
        takes control.

        Note: To facilitate the scheduler event processing thread to take
        control a 'sleep' call is invoked at the end of this method to
        avoid having to follow up with a 'sleep' call every time this method
        is invoked.

        Args:
            seconds (float): How many seconds to simulate passing in time.

        Raises:
            ValueError: If seconds is negative - can't go back in time
        """
        if float(seconds) < 0:
            raise ValueError("Cannot jump backwards in time")

        self.offset += seconds

        # have scheduler execute tasks that might be ready after this jump
        self._execute_pending_tasks()

    def _reset_scheduler(self):
        """ Reset our offset to 0 when the scheduler is reset """
        self.offset = 0
        super()._reset_scheduler()

    def _get_time(self):
        """ Overrides scheduler current time retrieval

        This time retrieval method is used when comparing against event_time
        """
        return super()._get_time() + self.offset

# Singleton reference to our jump ahead scheduler
JumpAheadScheduler = JumpAheadSchedulerRunner()



from nio.util.scheduler.scheduler import Scheduler


class Job(object):

    def __init__(self, target, delta, repeatable, *args, **kwargs):
        self._job = Scheduler.schedule_task(
            target, delta, repeatable, *args, **kwargs)

    def cancel(self):
        Scheduler.unschedule(self._job)



from enum import Enum

from nio.properties import BaseProperty
from nio.types import SelectType


class SelectProperty(BaseProperty):

    def __init__(self, enum, **kwargs):
        kwargs['enum'] = enum
        super().__init__(SelectType, **kwargs)
        self.description.update(self._get_description(**kwargs))

    def _get_description(self, **kwargs):
        """ Description needs to be json serializable """
        kwargs.update(self._prepare_options(kwargs['enum']))
        kwargs.update(self._prepare_default(**kwargs))
        kwargs['enum'] = kwargs['enum'].__name__
        return kwargs

    def _prepare_options(self, enum):
        # add internal object description
        options = {}
        for val in list(enum):
            options[val.name] = val.value
        return {"options": options}

    def _prepare_default(self, **kwargs):
        """ default in description should be serializable """
        default = kwargs.get('default', None)
        if isinstance(default, Enum):
            default = default.value
        return {"default": default}



"""
   Management signal for notifying changes to block or service status

"""
from nio.signal.management import ManagementSignal


class StatusSignal(ManagementSignal):

    """ An internal management signal to encapsulate status change information

    Args:
        status (RunnerStatus): The signal status.
        message (str): An optional explanation to go with the status change.

    """

    def __init__(self, status, message=None, **kwargs):
        super().__init__(**kwargs)
        self.status = status
        self.message = message


class ServiceStatusSignal(StatusSignal):

    """ A signal to use to notify service management signals.

    Args:
        status (RunnerStatus): The signal status.
        message (str): An optional explanation to go with the status change.
        service_name (str): The name of the service affected
        service_id (str): The identifier of the service affected

    """

    def __init__(self, status, message=None,
                 service_id=None, service_name=None, **kwargs):
        super().__init__(status, message, **kwargs)
        self.service_id = service_id
        self.service_name = service_name


class BlockStatusSignal(StatusSignal):

    """ A signal to use to notify block management signals.

    Args:
        status (RunnerStatus): The signal status.
        message (str): An optional explanation to go with the status change.
        block_name (str): The name of the block affected
        block_id (str): The identifier of the block affected
        service_id (str): The service the block is contained in
        service_name (str): The name of the service affected
    """

    def __init__(self, status, message=None,
                 block_name=None, block_id=None,
                 service_id=None, service_name="", **kwargs):
        super().__init__(status, message, **kwargs)
        self.block_name = block_name
        self.block_id = block_id
        self.service_id = service_id
        self.service_name = service_name



from nio.types.base import Type


class IntType(Type):

    @staticmethod
    def serialize(value, **kwargs):
        """ Convert a value to a JSON serializable value """
        return value

    @staticmethod
    def deserialize(value, **kwargs):
        """ Convert value to int """
        try:
            return int(value)
        except:
            raise TypeError("Unable to cast value to an int: {}".format(value))



from nio.modules.persistence import Persistence as PersistenceModule
from nio.modules.scheduler import Job
from nio.properties import TimeDeltaProperty, BoolProperty


class Persistence(object):

    """ A block mixin that provides persistence for some variables.

    To use, override the persisted_values function and define which variables
    on your class you wish to have persisted. The values should be strings
    that correspond to the variable names to be saved.

    """

    backup_interval = TimeDeltaProperty(
        visible=False, title='Backup Interval', default={"seconds": 60 * 60})
    load_from_persistence = BoolProperty(
        title='Load from Persistence?', default=True, advanced=True, order=100
    )

    def __init__(self):
        super().__init__()
        self._persistence = None
        self._backup_job = None

    def persisted_values(self):
        """ Return a list containing the values to be persisted.

        This function should be overriden in a Block that wishes to use
        persistence. Return a list of the names of the instance level variables
        to save.

        For example, if your block class has an instance level attribute called
        `_values` and you wish to save it, you would return this list:
            ["_values"]
        """
        return []

    def _load(self):
        """ Load the values from persistence

        Item is loaded from persistence using block id, once item
        is loaded, all persisted values are examined against loaded
        item, if it exists, such value is then set as an attribute
        in the block instance
        """
        self.logger.debug("Loading from persistence")
        # load whole item from persistence
        item = self._persistence.load(self.id(), default={})
        for persisted_var in self.persisted_values():
            if persisted_var in item:
                self.logger.debug("Loaded value {} for attribute {}".format(
                    item[persisted_var], persisted_var))
                # Set the loaded value to the attribute on this class
                setattr(self, persisted_var, item[persisted_var])

    def _save(self):
        """ Save the values to persistence
        """
        self.logger.debug("Saving to persistence")
        # generate item to be persisted by gathering all variables
        # to be persisted into a dictionary
        item = {persisted_var: getattr(self, persisted_var)
                for persisted_var in self.persisted_values()}
        # save generated dictionary under block's id
        self._persistence.save(item, self.id())

    def configure(self, context):
        super().configure(context)
        # Create a persistence object using the block's id
        self._persistence = PersistenceModule()
        if self.load_from_persistence():
            self._load()

    def start(self):
        super().start()

        # If we have configured a backup interval, set up the job to
        # periodically save to persistence
        if self.backup_interval().total_seconds() > 0:
            self._backup_job = Job(self._save, self.backup_interval(), True)

    def stop(self):
        if self._backup_job:
            self._backup_job.cancel()

        # Do one last save before stopping
        self._save()
        super().stop()



import heapq
from collections import namedtuple
from datetime import timedelta
from threading import Event, RLock
from time import monotonic
from uuid import uuid4

from nio.modules.module import ModuleNotInitialized
from nio.util.logging import get_nio_logger
from nio.util.runner import RunnerStatus, Runner
from nio.util.threading import spawn

QueueEvent = namedtuple('Event', 'time, id, target, frequency, args, kwargs')


class SchedulerRunner(Runner):

    def __init__(self):
        super().__init__()
        self._sched_min_delta = 0.1
        self._sched_resolution = 0.1
        self.logger = get_nio_logger("Custom Scheduler")
        self._queue = []
        self._queue_lock = RLock()
        self._stop_event = Event()
        self._events = dict()
        self._events_lock = RLock()
        self._process_events_thread = None
        # event used to wait for next task to execute and/or wait at scheduler
        # resolution
        self._sleep_interrupt_event = Event()

    def configure(self, context):
        # Load in the minimum delta and resolution from the config
        self._reset_scheduler()
        self._sched_min_delta = context.min_interval
        self._sched_resolution = context.resolution

    def _reset_scheduler(self):
        """ Reset the scheduler to the basic state.

        This method will be called whenever the scheduler needs to be
        restarted or start fresh. It will clear out the queue, reset the
        stop event, etc.
        """
        self._queue[:] = []
        # Set and then clear the event to trigger any needed stops
        self._stop_event.set()
        self._stop_event.clear()
        self._events.clear()
        if self._process_events_thread is not None:
            self._process_events_thread.join(self._sched_resolution)

    def schedule_task(self, target, delta, repeatable, *args, **kwargs):
        """ Add the given task to the Scheduler.

        Args:
            target (callable): The task to be scheduled.
            delta (timedelta): The scheduling interval. The scheduler will
                wait for this interval before running 'target' for the first
                time.
            repeatable (bool): When False, 'target' is run only once.
                Otherwise, it is run repeatedly at an interval defined by
                'delta'.
            args: Positional arguments to be passed to 'target'.
            kwargs: Keyword arguments to be passed to 'target'.

        Returns:
            job (Job): The Job object.

        """
        if self.status != RunnerStatus.started:
            raise ModuleNotInitialized("Scheduler module is not started")

        if not isinstance(delta, timedelta):
            raise AttributeError('delta must be of type: timedelta')

        delta = delta.total_seconds()
        if repeatable:
            # make sure delta is not smaller than minimum
            if delta < self._sched_min_delta:
                self.logger.warning(
                    "Scheduler delta of {} is invalid, minimum is {}".format(
                        delta, self._sched_min_delta))
                delta = self._sched_min_delta
            frequency = delta
        else:
            # when non repeatable, allow delta to be as small as user wishes
            # it to be
            frequency = 0

        event_id = uuid4().hex
        event = QueueEvent(
            self._get_time() + delta, event_id, target,
            frequency, args, kwargs)

        # add to queue
        with self._queue_lock:
            heapq.heappush(self._queue, event)

        # add to events
        with self._events_lock:
            self._events[event_id] = event

        return event_id

    def unschedule(self, job):
        """Remove a job from the scheduler.

        If the given job is not currently scheduled, this method
        has no effect.

        Args:
            job: The ID of the job to remove

        Returns:
            None

        """
        self.logger.debug("Un-scheduling %s" % job)
        # remove it from events dictionary
        event = None
        with self._events_lock:
            if job in self._events:
                event = self._events.pop(job)
        if event:
            try:
                with self._queue_lock:
                    if event in self._queue:
                        self._queue.remove(event)
                        heapq.heapify(self._queue)
                self.logger.debug('Success cancelling event')
                return True
            except Exception:
                self.logger.debug('Failure to remove event {0} from queue'
                                  ' while cancelling a job'.format(event))
        return False

    def stop(self):
        self._stop_event.set()
        # do not join indefinitely, allow a reasonable time
        self._process_events_thread.join(10 * self._sched_resolution)
        if self._process_events_thread.is_alive():
            self.logger.warning("Scheduler thread did not end properly, "
                                "it timed out")

    def start(self):
        self._process_events_thread = spawn(self._process_events)

    def _process_events(self):
        """ Process scheduled events

        Starts a loop that runs indefinitely until stop_event is set.
        Uses recommended-time returned from _execute_pending_tasks to wait
            for next pending tasks execution
        Any exception that may arise is logged while loop continues execution
        """

        while not self._stop_event.is_set():
            try:
                next_try_time = self._execute_pending_tasks()
                self._sleep_interrupt_event.wait(next_try_time)
            except Exception:
                # log any exception, do not leave loop
                self.logger.exception('Exception caught')

    def _execute_pending_tasks(self):
        """ Executes pending tasks

        This method will execute pending tasks, as soon as no task is ready for
        execution it will return.

        General characteristics:
            Scheduler tasks are launched asynchronously
            When not a single event is scheduled, method will return the
            resolution time, however, when events are present the next wait
            time is calculated as the minimum between scheduler's resolution
            and next event scheduled time.

        Returns:
            recommended time to wait before events are next considered
        """
        while not self._stop_event.is_set():
            with self._queue_lock:
                # is queue empty?
                if not self._queue:
                    # amount of time recommended to wait before trying again
                    return self._sched_resolution
                # have access to first event in queue
                event = heapq.heappop(self._queue)

            # check event's time to see if it is up for execution.
            event_time, event_id, target, frequency, args, kwargs = event
            # get time to compare event against
            now = self._get_time()
            if now < event_time:
                # event is in the future so push it back and recommend time
                # to wait before trying again
                with self._queue_lock:
                    heapq.heappush(self._queue, event)
                return min(event_time - now, self._sched_resolution)
            else:
                # time is up, execute
                try:
                    self.logger.debug("Executing: {0}".format(target))
                    # launch target task from a different thread thus
                    # making scheduler independent from task duration
                    spawn(target, *args, **kwargs)
                except Exception:
                    self.logger.exception('Calling: {0}'.format(target))

                with self._events_lock:
                    # before processing any further, make sure event has
                    # not been cancelled
                    if event_id in self._events:
                        # is it repeatable?
                        if frequency:
                            # reschedule it back, adding frequency to
                            # event time
                            event = QueueEvent(event_time + frequency,
                                               event_id,
                                               target,
                                               frequency,
                                               args, kwargs)
                            # housekeeping new event in
                            with self._queue_lock:
                                heapq.heappush(self._queue, event)
                            self._events[event_id] = event
                        else:
                            # remove event when not repeatable
                            del self._events[event_id]
                    else:
                        self.logger.debug("Event: {0} was cancelled".
                                          format(event_id))

    def _get_time(self):
        """ Time retrieval method to use when comparing against event time
        """
        # Use a clock that cannot go backwards.
        # This clock is not affected by system clock updates
        return monotonic()

# Singleton reference to a scheduler
Scheduler = SchedulerRunner()



from nio.modules.proxy import ModuleProxy


class Unauthorized(Exception):

    """ Unauthorized security Exception

    Exception triggered when a user is not authorized to perform a task
    """
    pass


class Authorizer(ModuleProxy):

    @classmethod
    def authorize(cls, user, task):
        """ Ensure a user can perform a task.

        Args:
            user (User): A user object, likely returned by the Authenticator
            task (SecureTask): A task this user can or cannot perform

        Returns:
            None: The method will pass successfully if the user is authorized

        Raises:
            Unauthorized: if the user cannot perform the specified task
        """
        raise NotImplementedError()

    @classmethod
    def is_authorized(cls, user, task):
        """ Finds out if user can perform a task.

        Args:
            user (User): A user object, likely returned by the Authenticator
            task (SecureTask): A task this user can or cannot perform

        Returns:
            True: if user can perform the specified task
            False: otherwise
        """
        try:
            cls.authorize(user, task)
            return True
        except Unauthorized:
            return False



""" discoverable and not_discoverable decorators

Classes marked as `discoverable` allow the system to identify them and register
them, while marking a class as  `not_discoverable` produces the opposite effect

Note: A class being marked `discoverable` will not mark its subclasses as
`discoverable`. Each subclass must be explicitly marked with the `discoverable`
decorator, the same applies for `not_discoverable`

To mark a class as not_discoverable, use the parameter-less decorator
`not_discoverable`::

    from nio import not_discoverable, Block

    @not_discoverable
    class MyBlock(Block):
        pass

"""


def discoverable(_class):
    """The decorator method to be called on the class object.

    This method will set the proper `discoverable` type to the class. It
    should return the class passed in, according to the decorator spec.
    """

    # Set the attribute to the class name, to prevent subclasses from also
    # being discoverable.
    setattr(_class, _get_discoverable_attribute(_class), True)
    return _class


def not_discoverable(_class):
    """The decorator method to be called on the class object.

    This method will set the proper `not discoverable` type to the class. It
    should return the class passed in, according to the decorator spec.
    """

    # Set the attribute to the class name, to prevent subclasses from also
    # being not discoverable.
    setattr(_class, _get_discoverable_attribute(_class), False)
    return _class


def is_class_discoverable(_class, default_discoverability=False):
    """ Returns true if the class is marked discoverable

    Note: discoverability does not extend to subclasses. Each subclass must
    explicitly be marked as discoverable

    Args:
        _class (class): The class in question, should be a class object
        default_discoverability: when attribute is not specified, this value
            will be provided.

    Returns:
        bool: True if the class is discoverable
    """
    return bool(getattr(_class, _get_discoverable_attribute(_class),
                        default_discoverability))


def _get_discoverable_attribute(_class):
    """ Get an attribute to set on a class to consider it discoverable """
    return "__{}_is_discoverable".format(_class.__name__)



from threading import Event
from time import sleep

from nio.util.threading import spawn


class ConditionWaiter(object):
    def __init__(self, event, callback, *args, sleep_step=0.001, **kwargs):
        self._event = event
        self._callback = callback
        self._sleep_step = sleep_step
        self._args = args
        self._kwargs = kwargs
        self._stopped = False

    def _check_condition(self):
        while not self._stopped:
            sleep(self._sleep_step)
            if self._callback(*self._args, **self._kwargs):
                self._event.set()
                break

    def start(self):
        self._stopped = False
        self._event.clear()
        spawn(self._check_condition)

    def stop(self):
        self._stopped = True


def ensure_condition(callback, *args,
                     sleep_step=0.001, max_wait_time=30, **kwargs):
    """ Waits for a True return from the callback specified

    Args:
        callback: callback to execute, when callback returns True
            the wait is over
        *args: callback arguments
        sleep_step: amount of time to sleep between condition checks
        max_wait_time: maximum amount of time to wait for condition
        **kwargs: callback arguments

    Returns:
        True: if callback at some point return True
        False: if waited maximum time allowed without callback ever
            returning True (CONDITION_EVENT_MAX_WAIT)

    """
    event = Event()
    condition = ConditionWaiter(event, callback, *args,
                                sleep_step=sleep_step, **kwargs)
    condition.start()
    result = event.wait(max_wait_time)
    condition.stop()

    return result



"""Terminal decorators and classes for defining inputs and outputs on blocks.

A terminal is an input or an output on a block that allows the block to handle
or notify different types of signals. Often a block will want to work with
different types of data and the service may want to take a different logic path
depending on the data. Terminals are perfect for this.

A good example is the Filter block. The Filter block will have two output
terminals, one for true and one for false. Depending on the evaluation of the
filter condition, the block will notify to either the true or false output
terminal. Then, the service is free to handle those different results however
it wants.
"""

from enum import Enum

# The ID that will be used to define the original default terminal
DEFAULT_TERMINAL = "__default_terminal_value"


class TerminalType(Enum):

    """Enum specifying the different types of terminals that are possible"""

    input = "input"
    output = "output"


class Terminal(object):

    """Decorator for declaring an input or output terminal on a block"""

    def __init__(self, terminal_type, terminal_id, default=False, label=None,
                 description=None, visible=True, order=0):
        """ Create a terminal with a given type and id

        Args:
            terminal_type (TerminalType): What type of terminal should this be
            terminal_id (str): The id of the terminal
            default (bool): Whether or not this terminal is the default
                terminal on the block
            label (str): The label to use for this terminal on a UI. Defaults
                to the terminal_id
            description (str): An optional, extended description for the
                terminal that can be used on a UI
            visible (bool): Whether or not the terminal should be visible on
                the UI. Defaults to True
            order (float): A numeric value used to determine the order this
                terminal should appear on the UI. Terminals will be ordered in
                ascending order

        Raises:
            TypeError: If terminal_type is not a valid TerminalType enum
        """

        if not isinstance(terminal_type, TerminalType):
            raise TypeError("Terminal type must be a TerminalType")
        self._type = terminal_type
        self.id = terminal_id
        self.default = default
        self.label = label if label is not None else terminal_id
        self.description = description if description is not None else ''
        self.visible = visible
        self.order = order

    def __call__(self, cls):
        """The decorator method to be called on the class object.

        This method will set the proper attribute to the class to save the
        terminal information. It should return the class passed in, according
        to the decorator spec.
        """
        # find out what entry we are saving on the class
        class_entry = self._get_terminals_entry(cls, self._type)

        # Get the attributes already on the class (or a new list, if not
        # present yet), and add our current terminal to the list.
        if not hasattr(cls, class_entry):
            setattr(cls, class_entry, list())
        attributes = getattr(cls, class_entry)
        attributes.append(self)
        return cls

    def get_description(self):
        """ Return a dictionary containing the description of this terminal """
        return {
            'type': self._type.value,
            'id': self.id,
            'label': self.label,
            'description': self.description,
            'default': self.default,
            'visible': self.visible,
            'order': self.order
        }

    @classmethod
    def get_terminals_on_class(cls, class_to_inspect, terminal_type,
                               order=True):
        """ Get a list of the unique terminals on a class of a certain type

        This method will recurse up the base classes and return all of the
        unique terminals that exist on it. Unique means that only one terminal
        will be included for each terminal_id. Duplicate IDs will be given to
        the class further down the MRO chain. This allows blocks to override
        certain properties on their parent blocks' terminals.

        Args:
            class_to_inspect (Block): A class that is a sub-class of Block that
                you want to get the terminals of
            terminal_type (TerminalType): What type of terminals to look for
            order (bool): Whether or not to sort the terminals based on the
                order attribute. If not, they will be sorted from first to
                last as the classes appear in the MRO

        Returns:
            list: A list of unique terminals on this block, sorted in order
                according to the terminal's order attribute

        Raises:
            TypeError: If terminal_type is not a valid TerminalType enum
        """
        if not isinstance(terminal_type, TerminalType):
            raise TypeError("Terminal type must be a TerminalType")

        # Start off our terminals set with our class's terminals
        class_entry = cls._get_terminals_entry(class_to_inspect, terminal_type)
        terminals = getattr(class_to_inspect, class_entry, list())

        # We also want to include the terminals of all super classes if they
        # don't already exist in ours
        for _class in class_to_inspect.__bases__:
            parent_terms = cls.get_terminals_on_class(_class, terminal_type)
            for parent_term in parent_terms:
                # If we don't already have a record of the parent terminal's
                # ID, then add the parent terminal to our output list
                if parent_term.id not in [term.id for term in terminals]:
                    terminals.append(parent_term)

        # Remove the DEFAULT_TERMINAL if it has other terminals defined
        if len(terminals) > 1:
            terminals = [t for t in terminals if t.id != DEFAULT_TERMINAL]

        if order:
            # Return the terminals sorted by their order
            return sorted(terminals, key=lambda t: t.order)
        else:
            return terminals

    @classmethod
    def get_default_terminal_on_class(cls, class_to_inspect, terminal_type):
        """ Gets the default terminal of a given type on a class

        If multiple terminals are declared as default, the one on the class
        that appears first in the MRO will be used. In other words, when
        iterating up the MRO, default terminals that are detected after a
        default terminal will be ignored.

        Args:
            class_to_inspect (Block): A class that is a sub-class of Block that
                you want to get the default terminal
            terminal_type (TerminalType): What type of terminals to look for

        Returns:
            Terminal: The default terminal on the class. None if no default
                exists

        Raises:
            TypeError: If terminal_type is not a valid TerminalType enum
        """
        if not isinstance(terminal_type, TerminalType):
            raise TypeError("Terminal type must be a TerminalType")

        # Iterate through the terminals ignoring their order attribute. We need
        # to do this so that we properly recurse up the MRO
        my_terms = cls.get_terminals_on_class(
            class_to_inspect, terminal_type, order=False)
        for terminal in my_terms:
            if terminal.default:
                return terminal
        return None

    @classmethod
    def _get_terminals_entry(cls, _class, terminal_type):
        """ Get the string for the attribute we should save to.

        This method will return a string indicating where subsequent attribute
        calls (i.e. getattr, setattr, hasattr) should be made on a class. The
        string will be unique for class names and terminal types. A different
        string needs to be returned for different classes so that it doesn't
        collide with subclasses or parent classes.

        Args:
            _class (cls): The class we will save the attribute on
            terminal_type (TerminalType): What type of terminal type we will
                be saving

        Returns:
            str: A string indicating where the attribute shoudl reside on the
                class
        """
        return "_{}_{}_attributes".format(_class.__name__, terminal_type.value)


class input(Terminal):

    """A decorator for an input terminal on a block"""

    def __init__(self, input_id, **kwargs):
        """Create an input terminal with a given id."""
        super().__init__(TerminalType.input, input_id, **kwargs)


class output(Terminal):

    """A decorator for an output terminal on a block"""

    def __init__(self, output_id, **kwargs):
        """Create an output terminal with a given id."""
        super().__init__(TerminalType.output, output_id, **kwargs)



import os
from configparser import RawConfigParser

from nio.project import Project, ConfigurationEntity, BlockEntity, ServiceEntity
from nio.project.entity import Entity
from nio.project.serializers.serializer import ProjectSerializer
from nio.util.codec import save_json, load_json, save_pickle, load_pickle
from nio.util.logging import get_nio_logger


class SerializationFormat(object):
    """ Holds serialization settings for a given format
    """
    def __init__(self, allowed_extensions, extension, load, save):
        self.allowed_extensions = allowed_extensions
        self.extension = extension
        self.load = load
        self.save = save


class FileSerializer(ProjectSerializer):

    links = [("logging", "conf", "etc/logging.json"),
             ("security", "users", "etc/users.json"),
             ("security", "permissions", "etc/permissions.json")]

    def __init__(self, project_path=None, conf_filename="nio.conf"):
        """ Initializes a config serializer instance

        Args:
            project_path (str): path to nio project location
            conf_filename (str): nio configuration file name
        """
        self.logger = get_nio_logger("FileSerializer")

        if project_path is None:
            project_path = os.getcwd()
        self._project_path = project_path
        self._conf_filename = conf_filename
        self.logger = get_nio_logger("File Project Serializer")

        # store in these format serializers specifics for each format
        self._json_format = SerializationFormat([".cfg", ".json"], ".cfg",
                                                load_json, save_json)
        self._pickle_format = SerializationFormat([".dat"], ".dat",
                                                  load_pickle, save_pickle)

    def deserialize(self, service_persistence=False):
        """ Deserializes a file n.io project to a Project instance.

        This method will read the files that this serializer class has been
        configured with and return a Project instance based on the structure
        and configuration of the files.

        Args:
            service_persistence (bool): Specifies if service persistence data 
                is de-serialized

        Returns:
            project (Project): A representation of this project as an instance
                of nio.project.Project
        """
        project = Project()

        if not os.path.isdir(self._project_path):
            raise ValueError("Folder {} does not exist".format(
                self._project_path))

        # change to project dir to account for relative paths
        os.chdir(self._project_path)

        # load nio conf first in case upcoming de-serializations depend
        # on anything we read from nio conf
        project.configuration = self._deserialize_nio_conf()

        core_p_folder = self._get_core_persistence_folder(project)

        # add blocks
        blocks_folder = os.path.join(core_p_folder, 'blocks')
        project.blocks = \
            self._deserialize_entities(blocks_folder,
                                       BlockEntity,
                                       self._json_format)

        services_folder = os.path.join(core_p_folder, 'services')
        # add services
        project.services = \
            self._deserialize_entities(services_folder,
                                       ServiceEntity,
                                       self._json_format)

        # deserialize core persistence
        project.core_persistence = \
            self._deserialize_persistence(core_p_folder,
                                          [blocks_folder, services_folder],
                                          self._json_format)

        if service_persistence:
            # deserialize service persistence
            service_p_folder = self._get_service_persistence_folder(project)
            project.service_persistence = \
                self._deserialize_persistence(service_p_folder, [],
                                              self._pickle_format)

        return project

    def serialize(self, project, service_persistence=False):
        """ Take the project instance and create the necessary files 

        Args:
            project (Project): configuration to serialize
            service_persistence (bool): Specifies if service persistence data 
                is serialized
        """

        # make sure target path exists
        if not os.path.isdir(self._project_path):
            os.makedirs(self._project_path)

        # change to project dir to account for relative paths
        os.chdir(self._project_path)

        # serialize configuration
        self._serialize_nio_conf(project.configuration)

        core_p_folder = self._get_core_persistence_folder(project)

        # serialize core persistence
        self._serialize_entities(project.blocks,
                                 os.path.join(core_p_folder, 'blocks'),
                                 self._json_format)

        # serialize services
        self._serialize_entities(project.services,
                                 os.path.join(core_p_folder, 'services'),
                                 self._json_format)

        self._serialize_entities(project.core_persistence, core_p_folder,
                                 self._json_format)

        if service_persistence:
            # serialize service persistence
            service_p_folder = self._get_service_persistence_folder(project)
            self._serialize_entities(project.service_persistence,
                                     service_p_folder,
                                     self._pickle_format)

    def _serialize_nio_conf(self, configuration):
        """ Serializes nio configuration settings to a file

        Note: Any existing configuration will be wiped out

        Args:
            configuration (dict): Configuration to serialize with format:
                {name: entity}
        """
        if not configuration:
            self.logger.info("No configuration items need to be serialized, "
                             "{} will not be affected".
                             format(self._conf_filename))
            return

        # create instance where settings will be stored and written from
        settings = RawConfigParser()
        # keep case sensitivity
        settings.optionxform = str

        # save special dict entries known to be saved as linked files
        serialized_entries = self._serialize_dict_entries(configuration)

        # write settings to parser
        for section, entity in configuration.items():
            # write section only if it has entries
            if entity.data:
                # check for section existence
                if not settings.has_section(section):
                    settings.add_section(section)
                # save every section option
                for option, value in entity.data.items():
                    # handle potential links since need to save the link
                    # instead of the value, (the link-value was not written into
                    # the configuration to preserve incoming configuration
                    # integrity)
                    if (section, option) in serialized_entries:
                        # override value with 'link' value
                        value = serialized_entries[(section, option)]
                    # save to settings resulting option value
                    settings.set(section, option, value)

        # figure out path to nio conf file
        nio_conf = os.path.join(self._project_path, self._conf_filename)
        # save actual config file
        with open(nio_conf, 'w') as fp:
            settings.write(fp)

    def _serialize_entities(self, entities, to_folder, ser_format):
        """ Serializes a list of entities to a folder.

        For example, this function can save in the 'services' folder
        a list of Service objects.

        Args:
            entities (dict): Entities to serialize with format: {name: entity}
            to_folder (str): The folder where the entities are saved
            ser_format (SerializationFormat): Serialization format
        """

        if not os.path.isdir(to_folder):
            os.makedirs(to_folder)

        for entity_name, entity in entities.items():
            if isinstance(entity, dict):
                # if entity is a dict, create a folder for it and
                # save each entry as an entity
                sub_folder = os.path.join(to_folder, entity_name)
                self._serialize_entities(entity, sub_folder, ser_format)
            elif not isinstance(entity, Entity):
                self.logger.warning('{} is not an entity, ignoring it'.
                                    format(entity_name))
                continue
            else:
                # create filename to serialize to
                filename = \
                    os.path.join(to_folder, "{}{}".format(entity_name,
                                                          ser_format.extension))
                try:
                    ser_format.save(filename, entity.data)
                except Exception:
                    # log exception and continue handling entities
                    self.logger.exception("Could not save entity data for: {}".
                                          format(entity_name))

    def _serialize_dict_entries(self, configuration):
        """ Handles entries known to contain its values as dictionaries

        Saves dictionary entries as stand-alone files and overrides actual
        entries to point to saved files.

        Args:
            configuration (dict): Configuration data with format:
                {name: entity}
        """

        serialized_entries = {}
        for (section, option, sub_path) in FileSerializer.links:
            if section not in configuration:
                # move to next if section is not in the incoming configuration
                continue
            data = configuration[section].data.get(option, None)
            if data is None:
                # move to next when option is not in incoming configuration
                continue

            # is data a potential link
            if isinstance(data, dict):
                # determine target filename
                filename = os.path.join(self._project_path, sub_path)
                # guarantee path to file
                path_to_file = os.path.dirname(filename)
                if not os.path.isdir(path_to_file):
                    os.makedirs(path_to_file)

                # save link file
                try:
                    save_json(filename, data)

                    # save entry so that it can be referenced
                    serialized_entries[(section, option)] = sub_path
                except Exception:
                    # log exception and continue handling entries
                    self.logger.exception(
                        "Could not save linked file for: {}.{}".
                        format(section, option))

        return serialized_entries

    def _deserialize_nio_conf(self):
        """ Parses and converts a nio conf file to a dictionary

        Entries that are known to point to a json file are expanded and
        a converted to a dictionary.

        Returns:
            A dictionary where each section in the config file is present
            under an entry
        """
        configuration = dict()

        # figure out path to nio conf file
        nio_conf = os.path.join(self._project_path, self._conf_filename)
        if not os.path.isfile(nio_conf):
            self.logger.info('There is no configuration data to deserialize')
            # return an empty configuration
            return configuration

        # read the contents of the conf file with the config parser
        settings = RawConfigParser()
        # keep case sensitivity
        settings.optionxform = str
        settings.read(nio_conf)

        # Populate our configuration dictionary
        for section in settings.sections():
            data = {
                option: settings.get(section, option)
                for option in settings.options(section)
            }
            configuration[section] = ConfigurationEntity(data=data)

        # special handling for entries known to be dictionaries
        self._deserialize_dict_entries(configuration)

        return configuration

    def _deserialize_entities(self, from_folder, entity_type, ser_format):
        """ Returns a list of entities loaded from a folder.

        For example, this function can look through the services/ folder
        to load a list of Service objects.

        Args:
            from_folder: The folder where the entity configurations are
            entity_type: A class to deserialize the configs in to
            ser_format (SerializationFormat): Serialization format

        Returns:
            entities (dict): A dict of entity type instances, keyed off of
                their filename basenames
        """

        entities = dict()
        if os.path.isdir(from_folder):
            for f in os.listdir(from_folder):
                key, entity = self._deserialize_entity(from_folder, f,
                                                       entity_type, ser_format)
                if key and entity:
                    entities[key] = entity

        return entities

    def _deserialize_persistence(self, folder, excluded, ser_format):
        """ Deserializes persistence data

        Args:
            folder (str): folder where persistence data resides
            excluded (list): list of sub-folders to exclude
            ser_format (SerializationFormat): Serialization format

        Returns:
            dictionary containing persistence data
        """
        data = {}
        if not os.path.isdir(folder):
            return data

        for f in os.listdir(folder):
            subdir = os.path.join(folder, f)
            if os.path.isdir(subdir):
                if subdir in excluded:
                    continue
                result = self._deserialize_persistence(subdir,
                                                       excluded,
                                                       ser_format)
                if result:
                    data[f] = result
            else:
                key, entity = self._deserialize_entity(folder, f,
                                                       Entity, ser_format)
                if key and entity:
                    data[key] = entity

        return data

    def _deserialize_entity(self, folder, file, entity_type, ser_format):
        """ Deserializes an entity

        Args:
            folder: folder where entity exists
            file: entity's file name
            entity_type: Entity type
            ser_format (SerializationFormat): Serialization format

        Returns:
            tuple containing (Key, Entity)
        """
        # grab extension and make sure it is one to process
        filename, extension = os.path.splitext(file)
        if extension in ser_format.allowed_extensions:
            try:
                # return tuple containing:
                # (filename without extension, actual entity instance)
                return filename, entity_type(
                    ser_format.load(os.path.join(folder, file)))
            except Exception:
                # handle json file errors
                self.logger.exception("Could not load entity data for: {}".
                                      format(filename))

        return None, None

    def _deserialize_dict_entries(self, sections):
        """ Handles entries known to contain its values as dictionaries

        Args:
            sections (dict): Sections loaded from configuration file

        Note: if entry points to a file, this file is loaded as a json file
        """

        for (section, option, default) in FileSerializer.links:
            if section not in sections:
                continue
            data = sections[section].data
            setting = data.get(option, default)
            # is it a link
            potential_link = os.path.join(self._project_path, setting)
            if os.path.isfile(potential_link):
                try:
                    # assume a json file
                    data[option] = load_json(potential_link)
                except:  # pragma: no cover
                    self.logger.exception(
                        "Could not load {} as json".
                        format(potential_link))  # pragma: no cover

    def _get_core_persistence_folder(self, project):
        try:
            core_p_folder = \
                project.configuration["persistence"].data["configuration_data"]
        except KeyError:
            # if entry is not defined, have 'etc' as fallback
            core_p_folder = os.path.join(self._project_path, "etc")

        return core_p_folder

    def _get_service_persistence_folder(self, project):
        try:
            service_p_folder = \
                project.configuration["persistence"].data["data"]
        except KeyError:
            # if entry is not defined, have 'etc/persist' as fallback
            service_p_folder = os.path.join(self._project_path,
                                            "etc", "persist")

        return service_p_folder



from sys import maxsize

from nio.modules.context import ModuleContext


class ModuleNotInitialized(Exception):
    """ Exception raised when module that has not been initialized

    Due to the asynchronous nature of some components in the system,
    a module might be accessed even when it has not been fully
    initialized, this exception should be raised for such cases.
    """
    pass


class Module(object):

    """ A base class for defining a module interface or implementation """

    default_version = "0.1.0"

    def initialize(self, context):
        """ Override this method to perform actions on initialization.

        Typically, during initialization, a module will want to proxy some
        class definitions and perform some additional setup items.

        Args:
            context (ModuleContext): Information pertaining to this module

        Returns:
            None - the return is ignored
        """
        pass

    def finalize(self):
        """ Override this method to perform actions on finalization.

        Typically, during finalization, a module will want to unproxy the
        classes it proxied when it initialized. There may also be some
        additional clean up items.

        Returns:
            None - the return is ignored
        """
        pass

    def get_module_order(self):
        """ Override this method to set the precedence for this module.

        This method should return an int specifying when this module should
        be initialized and finalized. Modules will be initialized in the
        order low-to-high and will be finalized in the reverse order. For
        example, a module with an order of 10 will be initialized before a
        module with an order of 20, but will be finalized after it.

        Defaults to a very large number, meaning it is initialized after the
        other modules.

        Returns:
            int: The order this module should be initialized
        """
        return maxsize

    def prepare_core_context(self):
        """ Prepare a ModuleContext in the core process

        Returns:
            ModuleContext: core module context
        """
        return ModuleContext()

    def prepare_service_context(self, service_context=None):
        """ Prepare a ModuleContext for the service process

        Takes a service context so it can know information about the service.
        This method will be called in the core process, but the context
        returned will be applied to the module in the service process.

        Args:
            service_context (ServiceContext): Service context (same context
                that is used when configuring a service)

        Returns:
            ModuleContext: service module context
        """
        return ModuleContext()

    def get_module_type(self):
        """ Provides module type

        Allows to provide a module type, note that this is not meant
        to be an actual implementation type.

        Returns:
            str: module type
        """
        raise NotImplementedError()

    def get_version(self):
        """ Provides module version

        Actual module implementations are encouraged to override this method.

        Returns:
            str: module version
        """
        return Module.default_version

    def get_details(self):
        """ Provides module details

        Actual module implementations may override this method and
        decide what details/properties they would like to expose.
        It is encouraged to call parent method and 'update' the result
        with own details

        """
        return {"version": self.get_version()}



from nio.properties import BaseProperty
from nio.properties import PropertyHolder
from nio.properties.util.object_type import ObjectType
from nio.types import ListType, Type


class ListProperty(BaseProperty):
    """ Defines a list property.

    List properties hold a list of object types which can contain
    properties themselves or can be nio types like IntType, etc.

    """

    def __init__(self, list_obj_type, **kwargs):
        """ Initializes the property.

        Args:
            list_obj_type (class): type of the elements to be contained
                in the list. Needs to be a subclass of nio.base.Type or
                a nio.properties.PropertyHolder
        """
        if issubclass(list_obj_type, PropertyHolder):
            kwargs['list_obj_type'] = ObjectType
            kwargs['obj_type'] = list_obj_type
        elif issubclass(list_obj_type, Type):
            kwargs['list_obj_type'] = list_obj_type
        else:
            raise TypeError("Specified list object type must be a "
                            "PropertyHolder or a nio Type")
        super().__init__(ListType, **kwargs)
        self.description.update(self._get_description(**kwargs))

    def _get_description(self, **kwargs):
        """ Description needs to be json serializable """
        kwargs.update(self._prepare_template(**kwargs))
        kwargs.update(self._prepare_default(**kwargs))
        kwargs['list_obj_type'] = kwargs['list_obj_type'].__name__
        if 'obj_type' in kwargs:
            kwargs['obj_type'] = kwargs['obj_type'].__name__
        return kwargs

    def _prepare_template(self, **kwargs):
        # add internal object description
        if "obj_type" in self.kwargs:
            # get description from PropertyHolder
            sub_description = self.kwargs["obj_type"]().get_description()
        else:
            # get class name from nio Type
            sub_description = self.kwargs["list_obj_type"].__name__
        return {"template": sub_description}

    def _prepare_default(self, **kwargs):
        """ default in description should be serializable """
        serializable_defaults = []
        defaults = kwargs.get('default', [])
        if self.is_expression(defaults) or self.is_env_var(defaults):
            # Don't mess with default if it's an expression or env var
            return {"default": defaults}
        for default in defaults:
            if isinstance(default, PropertyHolder):
                serializable_defaults.append(
                    default.get_serializable_defaults())
            else:
                serializable_defaults.append(default)
        return {"default": serializable_defaults}



class PermissionsError(Exception):
    pass


class PermissionsLine(object):

    """ Defines allowed permissions

    {
        # Resource level permissions:
        #  r: Can view a resource.
        #  w: Can edit a resource
        #  x: Can perform an execute operation like start/stop a resource.
    }

    Permissions can be specified as a combination of the above while keeping
    the order (see ALLOWED_PERMISSIONS).

    """

    ALLOWED_PERMISSIONS = ['r', 'w', 'x', 'rw', 'rx', 'wx', 'rwx']
    TARGETS = ["read", "write", "execute"]
    SHORTHAND = {"r": "read", "w": "write", "x": "execute"}

    def __init__(self, permissions=""):
        """ Initialized a permission line

        Args:
            permissions (string): Permissions to initialize line with
        """

        # allow uppercase definitions
        raw_perms = permissions.lower()
        if raw_perms and raw_perms not in PermissionsLine.ALLOWED_PERMISSIONS:
            raise PermissionsError("Invalid permission setting '{0}'".
                                   format(raw_perms))
        # assign permissions
        self.read = "r" in raw_perms
        self.write = "w" in raw_perms
        self.execute = "x" in raw_perms

    def __str__(self):
        r = "r" if self.read else ""
        r += "w" if self.write else ""
        r += "x" if self.execute else ""
        return r



from nio import __version__ as nio_version
from nio.modules.security import Authenticator
from nio.modules.security.access import set_user


class WebHandler(object):

    """ Web Handler root class

    A web handler is characterized by a route and a handler method (handle)
    where the application would be handling the given request.

    """

    def __init__(self, route):
        """ Create a new WebHandler instance at a given route

        Args:
            route (str): uri path

        """
        self.route = route

    def before_handler(self, request, response):
        """ Provides an entry point before the request is handled

        Performs request authentication by default.

        A given handler wishing not to have authentication or its own 
        authentication implementation would be able to do so by 
        overriding this method.

        Args:
            request (Request): A request from the web module used to determine
                the user who is making this request
            response (Response): Web response

        Returns:
            User: A reference to the user making the request.

        Raises:
            Unauthorized: If it cannot determine the user making the request
        """
        if request.get_method() == 'OPTIONS':
            # pre-flight requests do not contain security info
            return

        set_user(Authenticator.authenticate(request))

    def handle(self, request, response):
        """ Handles given web request

        Args:
            request (Request): Web request
            response (Response): Web response

        Returns:
            HTTP response
        """
        raise NotImplementedError()

    def after_handler(self, request, response):
        """ Provides an entry point after the request is handled

        Args:
            request (Request): Web request
            response (Response): Web response
        """
        # Add nio server tag header
        response.set_header("Server", "n.io v{}".format(nio_version))



from enum import Enum
import re

""" A valid version are three dotted with the format (major.minor.build)

Major and minor accept only digits and build accept characters (
accommodating for beta, release candidate (rc), etc)
"""
_version_regex = "^(?:(\d+)\.)?(?:(\d+)\.)?(\*|[a-zA-Z0-9]+)$"
_compiled_version_regex = re.compile(_version_regex)
_separators = "[.\-]+"


class VersionCheckResult(Enum):
    equal = 0
    newer = 1
    # not valid version results go below in decreasing numbers
    older = -1
    not_compatible = -2


class InvalidVersionFormat(Exception):
    pass


def is_version_valid(version):
    """ Allows to determine if a version string follows the version format

    Args:
        version: version to check against
    """
    return _compiled_version_regex.match(version) is not None


def compare_versions(a, b, separators=_separators, ignore_case=True):
    """ Compares two versions

    Args:
        a: pivot version
        b: version to compare against
        separators: version separator
        ignore_case: case sensitivity

    Returns:
        (VersionCheckResult): version comparison result
    """
    # convert versions to list
    a = _pre_process(a, separators, ignore_case)
    b = _pre_process(b, separators, ignore_case)

    # compare lists
    try:
        return VersionCheckResult((a > b) - (a < b))
    except:
        return VersionCheckResult.not_compatible


def get_major_version(version, separator="."):
    """ Figures out a major version from a regular version.

    It does so by constructing a version starting from the major digit and
    adding zero's to it.

    Args:
        version: version to check against
        separator: version separator

    Returns:
        major version
    """
    major = version.split(separator)[0]
    if not major.isdigit():
        raise ValueError("Major version must be numerical")
    return "{0}{1}0{1}0".format(major, separator)


def _pre_process(version, separators, ignore_case):
    """ Converts incoming version into a list based on separator

    Args:
        version: version to process
        separators: version separator
        ignore_case: case sensitivity

    For example:
        "1.2.1" => [1,2,1]
        "1.2.rc1" => [1,2,[rc,1]]
    """
    if ignore_case:
        version = version.lower()
    return [int(x) if x.isdigit()
            else [int(y) if y.isdigit()
                  else y for y in re.findall("\d+|[a-zA-Z]+", x)]
            for x in re.split(separators, version)]



"""
   NIO web support base class

"""
import json

import requests

from nio.modules.security.access import set_user, clear_user
from nio.modules.security.user import User
from nio.modules.web import WebEngine
from nio.testing.test_case import NIOTestCase


class NIOWebTestCase(NIOTestCase):

    """ Base Unit Test case class for Web related tests.

    Allows tests to concisely interact with the WebEngine
    and the requests module. Abstracts away some otherwise inevitable
    boilerplate.

    Add and remove servers as needed, specifying endpoint and config
    values. Check the status and contents of responses to requests
    made against those servers by name without ever touching the
    underlying response object.

    Example:
    server = self.add_server("foo", {"socket_port": 8182})
    server.add_handler(SomeHandler(<route>, <target>))
    self.start_engine()

    target = "http://%s:%d/foo" % (self.get_conf_host(), self.get_conf_port())
    self.assertResponseStatus('GET', target, 200)

    The above example adds a server and a handler, starts the CherryPy engine,
    and confirms that a GET request to the appropriate url returns HTTP Status
    Code 200.

    """

    def get_test_modules(self):
        """ Adds 'web' and 'security' to default modules """
        return super().get_test_modules() | {'web', 'security'}

    def setUp(self):
        super().setUp()
        self.servers = []
        set_user(User("tester"))

    def tearDown(self):
        clear_user()
        for server in list(self.servers):
            self.remove_server(server)
        super().tearDown()

    def start_engine(self):
        """ Starts the web engine """
        WebEngine.start()

    def add_server(self, port, config=None, host="127.0.0.1", auto_start=True):
        """ Adds and starts a web server (if auto_start is True)

        Args:
            port (int): web server port
            config (dict): server configuration parameters
            host (str): web server host
            auto_start (bool): if True server is started

        """
        web_server = WebEngine.add_server(port, host,
                                          config if config is not None else {})
        # make sure server is not already registered
        if web_server not in self.servers:
            self.servers.append(web_server)
        if auto_start:
            web_server.start(config)
        return web_server

    def remove_server(self, server):
        """ Removes a web server

        Args:
            server: web server

        """
        if server in self.servers:
            self.servers.remove(server)
        WebEngine.remove_server(server)

    def get_conf_host(self):
        """ Returns default host """
        return '127.0.0.1'

    def get_conf_port(self):
        """ Returns default port """
        return 8181

    def get_response(self, method, url, **kwargs):
        """ Get a response object for a given resource

        Args:
            method (str): e.g. GET, POST, etc.
            url (str): Complete target for the request
            kwargs (dict): Payload for POST/PUT

        """
        return requests.request(method, url, **kwargs)

    def assertResponseStatus(self, method, url, status, **kwargs):
        """ Checks that the specified request returns the specified status.

        Args:
            method (str): e.g. GET, POST, etc.
            url (str): Complete target for the request
            status (int): The expected status for the response.
            kwargs (dict): Payload for POST/PUT

        """
        response = self.get_response(method, url, **kwargs)
        self.assertEqual(response.status_code, status)

    def assertResponseJSON(self, method, url, key, value=None, data=None):
        """ Assert that the specified response returns a JSON with a value

        Args:
            method (str): e.g. GET, POST, etc.
            url (str): Complete target for the request
            key (str): The JSON response is expected to contain this key.
            value (str): If set, asserts that (key, value) is in the response.
            data (str): Payload for POST/PUT requests.

        """
        response = self.get_response(method, url, data=data)
        data = json.loads(response.text)
        if value is not None:
            self.assertEqual(data[key], value)
        else:
            self.assertIn(key, data)

    def assertResponseJSONValue(self, method, url, value, data=None,
                                status=200):
        """ Assert that the specified response returns a JSON matching a value

        Args:
            method (str): e.g. GET, POST, etc.
            url (str): Complete target for the request
            value (dict): Expected Http Response
            data (str): Payload for POST/PUT requests.

        """
        response = self.get_response(method, url, data=data)
        self.assertEqual(response.status_code, status)
        data = json.loads(response.text)
        self.assertDictEqual(data, value)



from nio.modules.context import ModuleContext
from nio.testing.modules.settings import Settings
from nio.modules.settings.module import SettingsModule


class TestingSettingsModule(SettingsModule):

    def initialize(self, context):
        super().initialize(context)
        if not context.in_service:
            self.proxy_settings_class(Settings)
        else:
            # Don't proxy Settings, we want NotImplementedError to be raised
            # inside the service process
            pass

    def finalize(self):
        super().finalize()

    def prepare_core_context(self):
        context = ModuleContext()
        context.in_service = False
        return context

    def prepare_service_context(self, service_context=None):
        context = ModuleContext()
        context.in_service = True
        return context



from nio.block.base import Block
from nio.block.generator_block import GeneratorBlock
from nio.block.terminator_block import TerminatorBlock
from nio.signal.base import Signal
from nio.modules.module import Module
from nio.util.discovery import discoverable

__version__ = '3.1.1'



from nio.signal.base import Signal


class LogSignal(Signal):

    """ A log signal to encapsulate log record information

    """

    def __init__(self,
                 time,
                 context,
                 level,
                 message,
                 filename,
                 function,
                 line
                 ):
        """ Create a new LogSignal instance.

        Args:
            time: logging instruction time
            context: logging instruction context
            level: logging level
            message: logging message
            filename: filename where logging instruction resides
            function: function where logging instruction resides
            line: line number where logging instruction resides

        """
        super().__init__({"time": time,
                          "context": context,
                          "level": level,
                          "message": message,
                          "filename": filename,
                          "function": function,
                          "line": line
                          })



from nio.command.params.base import Parameter
from nio.command.params.string import StringParameter


class InvalidCommandArg(Exception):
    pass


class MissingCommandArg(Exception):
    pass


class Command(object):

    """Command to be used in CommandHolders (Blocks and Services).

    Args:
        id (str): The id of the command. Should be the same as the name of
            the corresponding method.
        params (Properties): List of Property Params that are part of the
            command
        title (str): An optionally more detailed description of the
            command. If left blank, title defaults to id.
        method (str): The id of the real method exposed by this command if
            different than id

    """

    def __init__(self, id, *params, title=None, method=None):
        self._id = id
        self._parameters = []
        for p in params:
            self._add_parameter(p)
        self._title = title or id
        self._method = method or id

    @property
    def id(self):
        return self._id

    @property
    def title(self):
        return self._title

    @property
    def parameters(self):
        return self._parameters

    @property
    def method(self):
        return self._method

    def get_description(self):
        """Returns a description for a command instance.

        This method is used iteratively in CommandHolder.get_description.

        Returns:
            description (dict): the title of the command and descriptions
                of its parameters
        """
        description = {"title": self.title}
        params = {}
        for p in self._parameters:
            params[p.name] = p.get_description()
        description["params"] = params
        return description

    def _add_parameter(self, param):
        if isinstance(param, Parameter):
            self._parameters.append(param)
        else:
            raise RuntimeError("Invalid command parameter type: '{}'".format(
                getattr(param, 'name', param)))

    def collect_arguments(self, args):
        """Format passed arguments into parameter args and kwargs.

        Assign values to the command's parameters. This is done to
        provide safe conversion of command arguments from strings (they
        will likely arrive as strings in an HTTP request body) and to
        ensure that arguments are of the correct types.

        This method converts each argument to the appropriate type per
        each parameter. Collection fails if, for any parameter, allow_none
        is false and the type conversion returns None.

        Args:
            args (dict): The arguments.

        Returns:
            args (list): The list of arguments, converted and type
                checked.
        """
        result = []
        result_kwargs = {}
        if isinstance(args, dict):

            # Find and convert each param needed
            for p in self._parameters:
                raw_val = args.get(p.name) or p.default
                if raw_val is not None or p.allow_none:
                    val = p.deserialize(raw_val)
                    result.append(val)
                else:
                    raise MissingCommandArg(
                        "Command {}: missing argument {}".format(
                            self.id, p.name))
            # Now that we have all params that were not passed but have
            # default value covered, lets check for unrecognized param passed
            invalid_args = [arg for arg in args if arg not in
                            [p.name for p in self._parameters]]
            # Treat all extra args in the passed dict as kwargs
            for arg in invalid_args:
                result_kwargs[arg] = args.get(arg)
        else:
            # Try to use whole arg as the param if the method only
            # needs one
            if len(self._parameters) == 1 and \
                    self._parameters[0].__class__ == StringParameter:
                result.append(args)
            else:
                raise InvalidCommandArg(
                    "Invalid arguments (dict expected): {0}".format(args))
        return result, result_kwargs



from types import BuiltinMethodType, MethodType, FunctionType


class ClassAttributes(object):

    @classmethod
    def is_attr(cls, attr_value):
        """ Determine if an attribute value is a 'data' attribute.

        Args:
            attr_value: Attribute value to check against

        Returns:
            True if value is a 'data' attribute, False otherwise
        """
        if (attr_value is BuiltinMethodType or
                isinstance(attr_value, BuiltinMethodType) or
                isinstance(attr_value, type(object().__hash__)) or
                isinstance(attr_value, MethodType) or
                isinstance(attr_value, FunctionType)):
            return False
        return True



import logging

from nio.util.logging.handlers.publisher.cache import LogCache


class CacheFilter(logging.Filter):

    def __init__(self, name='', expire_interval=1):
        """ Create a new CacheFilter instance.

        Args:
            name: see parent class logging.Filter
            expire_interval (seconds): specifies how long an entry will be
             kept in the cache

        """
        super().__init__(name=name)
        self._log_cache = LogCache(expire_interval)

    def filter(self, record):
        """  Filters the record before being logged

        If the record is determined to be in the cache, it is filtered
        out.

        Args:
            record: record being filtered

        Returns:
            True if record should be logged, False otherwise.
        """
        return not self._log_cache.process_record(record)



from nio.command.params.base import Parameter
from nio.types import IntType


class IntParameter(Parameter):

    def __init__(self, name, **kwargs):
        super().__init__(IntType, name, **kwargs)



from nio.modules.proxy import ModuleProxy


class WebEngine(ModuleProxy):

    @classmethod
    def add_server(cls, port, host=None, configuration=None):
        """ Creates and returns a web server

        Args:
            port (int): Port listening for requests
            host (str): Web server host address. Should default to the current
                host of the server.
            configuration (dict): Contains additional server configuration,
                might contain additional ssl settings such as:
                ssl_certificate, ssl_private_key, ssl_certificate_chain

        Returns:
            The server if it already exists, or a new instantiated server
        """
        raise NotImplementedError()

    @classmethod
    def remove_server(cls, server):
        """ Stops and remove a server

        Args:
            server (WebServer): Server to remove
        """
        raise NotImplementedError()

    @classmethod
    def start(cls):
        """ Starts Web engine

        Starts a web engine based on previously set configuration
        """
        raise NotImplementedError()

    @classmethod
    def stop(cls):
        """ Stops engine and running Web Servers """
        raise NotImplementedError()

    @classmethod
    def block(cls):
        """ Blocks the Web engine

        This method is normally called from the main thread to block
        execution since likely each request will be handled by a separate
        thread.

        When stopping the engine, it is expected for it wait for all currently
        executing requests/threads to terminate.

        """
        raise NotImplementedError()



from threading import RLock, Event
from time import sleep, monotonic

from nio.modules.communication.publisher import Publisher
from nio.modules.module import ModuleNotInitialized
from nio.modules.proxy import ProxyNotProxied
from nio.util.threading import spawn


class PublisherNotReadyException(Exception):
    """ Exception raised when publisher is not ready within specified time
    """
    pass


class PublisherProxy(object):

    _publisher = None
    _publisher_lock = RLock()
    # establish that publisher is ready for use
    _publisher_ready_event = Event()
    # many loggers get initialized however, only once the proxy is initialized
    _initialized = False
    # publishing criteria
    _topic = "nio_logging"
    # Maximum time to wait for Publisher to be ready
    _max_publisher_ready_time = 5
    # Interval between sleeps waiting for Publisher to be ready
    _publisher_ready_wait_interval_time = 0.1
    # thread created to check asynchronously on publisher readiness
    _open_thread = None

    @classmethod
    def init(cls, topic,
             max_publisher_ready_time, publisher_ready_wait_interval_time):
        """ Initializes the proxy

        Args:
            topic (str): topic to use when publishing log messages
            max_publisher_ready_time (float): maximum time to wait for
                publisher to be ready
            publisher_ready_wait_interval_time (float): interval in seconds to
                use when waiting for publisher to be ready
        """
        if not cls._initialized:
            # proxy initialization happens only once
            cls._topic = topic
            cls._max_publisher_ready_time = max_publisher_ready_time
            cls._publisher_ready_wait_interval_time = \
                publisher_ready_wait_interval_time
            cls._open_thread = spawn(cls._create_publisher)
            cls._initialized = True

    @classmethod
    def publish(cls, signals):
        """ Publishes logging signals.

        This method publishes the signal when publisher has been setup
        successfully.

        It might happen that a call to publish 'logging' signals comes before
        publisher is ready, this is due to the logging module being initialized
        before communication module, in which case the signals are ignored or
        not published

        Note: establishing a mechanism waiting for publisher to be ready during
        this 'publish' call has its own drawbacks due to the fact that the
        original logging instructions might hold back the execution that allows
        a publisher to be ready in the first place (see _create_publisher)

        Args:
            signals: signals to publish

        """
        # publish when all conditions are met
        if cls._publisher_ready_event.is_set():
            with cls._publisher_lock:
                cls._publisher.send(signals)
        elif isinstance(cls._open_thread.nio_exception, Exception):
            # even though the thread is likely terminated, it still holds
            # the exception if there was one, so do not let the exception
            # to go unnoticed if still trying to publish
            raise cls._open_thread.nio_exception

    @classmethod
    def close(cls):
        """ Closes publisher by releasing all resources """

        if cls._publisher:
            try:
                cls._publisher.close()
            # catch NotImplementedError since when stopping a service
            # the 'Publisher' interface is eventually un-proxied
            except NotImplementedError:
                pass
            finally:
                cls._publisher = None

        cls._publisher_ready_event.clear()
        cls._initialized = False

    @classmethod
    def _create_publisher(cls):
        """ Attempts to create publisher.

        Keeps on trying to create publisher, which will happen successfully
        once communication module is initialized
        """
        end_time = monotonic() + cls._max_publisher_ready_time
        while not cls._publisher_ready_event.is_set():
            try:
                cls._publisher = Publisher(topic=cls._topic)
                cls._publisher.open()
                cls._publisher_ready_event.set()
            except (ProxyNotProxied, ModuleNotInitialized, NotImplementedError):
                if monotonic() >= end_time:
                    raise PublisherNotReadyException(
                        "Maximum time for publisher to be ready elapsed")
                sleep(cls._publisher_ready_wait_interval_time)



from nio.testing.modules.security.authenticator import Authenticator
from nio.testing.modules.security.authorizer import Authorizer
from nio.modules.security.module import SecurityModule


class TestingSecurityModule(SecurityModule):

    def initialize(self, context):
        super().initialize(context)
        self.proxy_authenticator_class(Authenticator)
        self.proxy_authorizer_class(Authorizer)



class Type(object):
    """ Base type for property and parameters """

    def __init__(self):
        # Type is a static class
        raise RuntimeError("A Type should never be instantiated")

    @staticmethod
    def serialize(value, **kwargs):
        return value

    @staticmethod
    def deserialize(value, **kwargs):
        return value



from nio.modules.web.handlers.web import WebHandler


class RESTHandler(WebHandler):

    def __init__(self, route):
        """ Create a new RESTHandler instance.

        Args:
            route: uri path

        """
        super().__init__(route)

    def is_collection(self):
        """ Specifies if the request result is a collection of items

        Returns:
            True if a collection is returned, False otherwise

        """
        return False

    def on_get(self, request, response):
        """ Get request handler

        The HTTP GET method is used to **read** (or retrieve) a
        representation of a resource

        Args:
            request: web request
            response: web response

        Returns:
            Web result

        """
        raise NotImplementedError()

    def on_post(self, request, response):
        """ Post request handler

        The POST verb is most-often utilized to **create** new resources

        Args:
            request: web request
            response: web response

        Returns:
            Web result

        """
        raise NotImplementedError()

    def on_put(self, request, response):
        """ Put request handler

        PUT is most-often utilized for **update** capabilities,
        PUT-ing to a known resource URI with the request body containing
        the newly-updated representation of the original resource.

        Args:
            request: web request
            response: web response

        Returns:
            Web result

        """
        raise NotImplementedError()

    def on_delete(self, request, response):
        """ Delete request handler

        Used to **delete** a resource identified by a URI.

        Args:
            request: web request
            response: web response

        Returns:
            Web result

        """
        raise NotImplementedError()

    def on_options(self, request, response):
        """ Options request handler, respond 200 by default

        We respond 200 by default for default CORS response. You are allowed
        to override with your own implementation.

        Args:
            request: web request
            response: web response

        Returns:
            Web result

        """

        response.set_status(200)
        response.set_body("")



""" An object that keeps track of modules and configurations.

A ModuleInitializer will have a list of module (nio.modules.module.Module)
objects as well as what configuration contexts they should be initialized
with. When instructed to initialize, the ModuleInitializer will determine
the proper order to initialize the modules based on the module class
definitions.
"""
from nio.util.logging import get_nio_logger
from nio.modules.proxy import ProxyNotProxied, ProxyAlreadyProxied
from nio.modules.module import Module


class ModuleInitializer(object):

    """ A class that proxies and unproxies modules """

    def __init__(self):
        super().__init__()
        self._registered_modules = []
        self._initialized_modules = []

    def initialize(self, safe=False):
        """ Initialize and proxy any registered module proxy implementations.

        This function will initialize any module implementations that have been
        registered with this ModuleInitializer.

        Args:
            safe (bool): True if you want the initialize to ignore modules
                trying to implement an already implemented interface.
                Defaults to False
        """
        for module in sorted(self._registered_modules,
                             key=lambda mod: mod.get_module_order()):

            # make sure module is not already initialized
            if module in self._initialized_modules:
                continue

            self._initialize_module(module, self.get_context(module), safe)

    def register_module(self, module):
        """ Register a module and context to use for its initialization

        Args:
            module (Module): A nio module implementation

        Raises:
            TypeError: if the implementation class is not a class
        """
        if not isinstance(module, Module):
            raise TypeError("Registered module must be an instance of Module")
        self._registered_modules.append(module)

    def finalize(self, safe=False):
        """ De-register and unproxy the registered modules on this initializer

        This will finalize the modules in the reverse order that they
        were proxied.

        Args:
            safe (bool): True if you want the finalize to succeed even if
                the modules are already unproxied. Defaults to False
        """
        for module in reversed(self._initialized_modules):
            if safe:
                try:
                    module.finalize()
                except ProxyNotProxied:
                    get_nio_logger('ModuleInitializer').warning(
                        "Ignoring ProxyNotProxied while finalizing {}".format(
                            module.__class__.__name__))
            else:
                module.finalize()
        self._initialized_modules[:] = []

    def _initialize_module(self, module, context, safe):
        try:
            module.initialize(context)
        except ProxyAlreadyProxied:
            get_nio_logger('ModuleInitializer').warning(
                "Interface implemented by module '{}' is already proxied".
                format(module.__class__.__name__))
            if safe:
                # ignore a module trying to re-proxy when in safe mode
                return
            else:
                raise

        get_nio_logger('ModuleInitializer').info(
            "Module {} is initialized".format(module.__class__.__name__))

        # Add this module to the list of modules that have been proxied
        self._initialized_modules.append(module)

    def get_context(self, module):
        raise NotImplementedError()



from nio.modules.module import Module
from nio.modules.scheduler.job import Job


class SchedulerModule(Module):

    def proxy_job_class(self, job_class):
        Job.proxy(job_class)

    def finalize(self):
        Job.unproxy()
        super().finalize()

    def get_module_order(self):
        return 5

    def get_module_type(self):
        return "scheduler"



from datetime import timedelta
from threading import RLock

from nio.modules.scheduler.job import Job


class Cache(object):
    """ Utility class to cache and expire items given a certain duration

    There is no distinction as to what kind of items can be cached in a
    given instance, therefore, items of different types can be combined within
    the same cache.
    """

    def __init__(self, duration):
        """ Create a new Cache instance.

        Args:
            duration: default duration in seconds to use for items,
                it can be overriden when adding an item
        """
        self._duration = duration
        self._cache = dict()
        self._cache_lock = RLock()

    def add(self, key, item, duration=None):
        """ Adds an item to the cache

        When an item is added with a positive duration, a job is scheduled
        for its removal at given duration in seconds

        Args:
            key: item key
            item: item value
            duration: if specified item will be kept for this duration in
                seconds, otherwise instance-wide duration is used
        """
        if duration is None:
            # use instance-wide duration when specific duration is not provided
            duration = self._duration

        # is key already in cache
        with self._cache_lock:
            if key in self._cache:
                # cancel job when overriding
                _, job = self._cache.get(key)
                job.cancel()

            if duration > 0:
                # schedule expiration
                job = Job(self._remove_item, timedelta(seconds=duration),
                          False, key)

                # save item in cache
                self._cache[key] = (item, job)

    def get(self, key, default=None):
        """ Gets item from cache

        Args:
            key: item key
            default: value to return when key is not in cache

        Returns:
            item if found, default otherwise
        """
        with self._cache_lock:
            try:
                item, _ = self._cache.get(key)
                return item
            except:
                return default

    def _remove_item(self, key):
        """ Removes item from cache

        This method is triggered when item expires while being called from
        system scheduler.

        Args:
            key: item key
        """
        with self._cache_lock:
            if key in self._cache:
                del self._cache[key]



from .base import Type
from .bool import BoolType
from .dict import DictType
from .file import FileType
from .float import FloatType
from .int import IntType
from .list import ListType
from .select import SelectType
from .string import StringType
from .timedelta import TimeDeltaType



import re

from setuptools import setup, find_packages

with open('nio/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

def readme():
    with open('README.md') as f:
        return f.read()

setup(
    name='nio',
    description='A framework for building blocks and modules for the nio Platform',
    long_description=readme(),
    keywords=['nio'],
    version=version,
    packages=find_packages(
        exclude=['*.tests', '*.tests.*', 'tests.*', 'tests']),

    install_requires=['safepickle>=0.1.0'],

    tests_require=[
        'requests>=2.3.0'
    ],
	
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
    ],

    license='Apache 2.0',
    author='niolabs',
    author_email='support@n.io',
    url='https://github.com/niolabs/nio'
)



class TestingRequest(object):

    def get_body(self):
        return None

    def get_identifier(self):
        return None

    def get_params(self):
        return {}

    def get_headers(self):
        return {}

    def get_method(self):
        return "UNKNOWN"


class TestingResponse(object):

    def set_header(self, header_name, header_value):
        pass

    def set_status(self, status, message=None):
        pass

    def set_body(self, body):
        pass



from enum import Enum
import logging


class LogLevel(Enum):
    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    NOTSET = logging.NOTSET



from nio.command.base import Command
from nio.command.holder import CommandHolder


def command(*args, **kwargs):
    """Command decorator to add Commands to CommandHolders."""
    def wrap_command(cls):
        if not issubclass(cls, CommandHolder):
            raise TypeError("Must place command decorator on a CommandHolder")
        cls.add_command(Command(*args, **kwargs))
        return cls
    return wrap_command



from nio.modules.proxy import ModuleProxy


class Job(ModuleProxy):

    """A scheduled job in the scheduler module """

    def __init__(self, target, delta, repeatable, *args, **kwargs):
        """ Create a new job instance.

        Args:
            target (callable): The task to be scheduled.
            delta (timedelta): The scheduling interval. The scheduler will
                wait for this interval before running 'target' for the first
                time.
            repeatable (bool): When False, 'target' is run only once.
                Otherwise, it is run repeatedly at an interval defined by
                'delta'.
            args: Positional arguments to be passed to 'target'.
            kwargs: Keyword arguments to be passed to 'target'.

        """
        super().__init__(target, delta, repeatable, *args, **kwargs)

    def cancel(self):
        """ Unschedule this task.

        """
        raise NotImplementedError()



from nio.properties import BaseProperty
from nio.types.base import Type


class Property(BaseProperty):

    """ A property that can assume any type """

    def __init__(self, **kwargs):
        super().__init__(Type, **kwargs)



from .base import BaseProperty
from .bool import BoolProperty
from .holder import PropertyHolder
from .int import IntProperty
from .float import FloatProperty
from .list import ListProperty
from .object import ObjectProperty
from .property import Property
from .select import SelectProperty
from .string import StringProperty
from .timedelta import TimeDeltaProperty
from .version import VersionProperty
from .file import FileProperty



from nio.modules.security.authorizer import Authorizer, Unauthorized
from nio.modules.security.authenticator import Authenticator
from nio.modules.security.user import User
from nio.modules.security.task import SecureTask



from threading import Thread

from nio.util.logging import get_nio_logger


class NIOThread(Thread):

    def __init__(self, **kwargs):
        """ Create a new NIOThread instance.

        Args:
            kwargs: keyword arguments

        """
        super().__init__(**kwargs)
        self.logger = get_nio_logger("NIOThread")
        self.nio_result = None
        self.nio_exception = None

    def run(self):
        """ Overrides task execution

        Execution is overriden with the purpose of providing feedback when
        the result of the asynchronous task is of interest to caller

        Target execution result is saved in 'nio_result'

        Likewise, if an exception is thrown by target, it is captured and saved
        in 'nio_exception'

        """
        try:
            self.nio_result = self._target(*self._args, **self._kwargs)
        except BaseException as e:
            # first element in _args tuple contains the method to call
            self.logger.exception("Executing: {0}".
                                  format(self._target.__name__))
            self.nio_exception = e

    def join(self, timeout=None):
        """ Overrides join execution

        Args:
            timeout: join timeout

        Returns:
            target execution result

        """
        super().join(timeout)
        if self.nio_exception:
            raise self.nio_exception
        return self.nio_result

    def __repr__(self):
        return "{}, nio_result: {}, nio_exception: {}".format(
            super().__repr__(), self.nio_result, self.nio_exception)



from nio.util.threading.thread import NIOThread


def spawn(target, *args, **kwargs):
    """ Executes given target in a new thread

    This method creates a thread invoking the target function specified

    If the target function happens to eventually invoke a longtime running
    function executing in C, the GIL happens to hijack control and would not
    allow the launching thread to continue until the function ends, such
    behaviour can be mitigated by wrapping 'target' in a new function where
    the first instruction is a 'sleep' with a very low number of seconds
    (0.00001 is recommended)

    Args:
        target: method or function to execute
        *args: positional arguments in 'target'
        **kwargs: keyword arguments in 'target'

    Returns:
        newly created thread

    """
    t = NIOThread(target=target, args=args, kwargs=kwargs)
    t.daemon = True
    t.start()
    return t



from nio.modules.module import Module
from nio.modules.settings import Settings


class SettingsModule(Module):

    def proxy_settings_class(self, settings_class):
        Settings.proxy(settings_class)

    def finalize(self):
        Settings.unproxy()
        super().finalize()

    def get_module_order(self):
        return 1

    def get_module_type(self):
        return "settings"



from nio.modules.web.exceptions import HTTPError
from nio.modules.web.exceptions import HTTPNotImplemented
from nio.modules.web.exceptions import HTTPUnauthorized
from nio.modules.web.exceptions import HTTPNotFound

from nio.modules.web.request import Request
from nio.modules.web.response import Response



from copy import copy
from collections import defaultdict, Hashable
from nio.properties import Property
from nio.util.ensure_types import ensure_list
from nio.command import command
from nio.command.holder import CommandHolder


@command("groups", method='_groups_command')
class GroupBy(CommandHolder):

    """ Use this base class to extend block functionality to groups of signals

    Use NIO's native expression syntax to configure a value by which to
    group signals. In your block's 'process_signals' method, pass the
    list of signals and a target method to self.for_each_group. The
    signals will be grouped automatically and the target method called
    on each group in turn.

    If you'd like to map a function over all the groups outside of
    process_signals, just call self.for_each_group without passing any
    signals.

    You can also make use of the default process_signals implmentation of
    grouping all incoming signals and processing them independently by not
    overriding process_signals but instead overriding process_group_signals.

    Properties:
        group_by: The expression by which signals will be grouped.

    """

    group_by = Property(title="Group By", default=None, order=100, allow_none=True)

    def __init__(self):
        super().__init__()
        self._groups = set()

    def for_each_group(self, target, signals=None, *args, **kwargs):
        """ Execute a function once for every group

        If signals are provided, the target function will only be called
        for groups that appear in the provided signals. If signals are omitted,
        the function will be called for every group that has ever been
        processed in this block.

        Note that the required template for 'target' depends on
        whether or not signals are provided for grouping.

        Args:
            target (callable): The target function.
            signals (list): The list of signals to be grouped. If
                this is left off, the list of recorded groups is
                used.
            args, kwargs : Additional arguments to be passed to target

        Returns:
            list(Signal): A list of the return signals of each call to the
                target function. If the target function returns a list, the
                output list will be a concatenation of results.
                So [Signal, Signal, Signal] not [[Signal, Signal],[Signal]]]
        """
        output = []
        # if there are no signals, assume that the target function has
        # only one parameter, the group key
        if signals is None:
            # We are going to map each group key over the target function
            # and produce the output along the way.
            # We need to send a copy of self._groups so that in case the target
            # function alters the groups list it doesn't affect iteration
            for group in copy(self._groups):
                result = target(group, *args, **kwargs)
                if result:
                    output.extend(ensure_list(result))

        # otherwise, assume that the target function has two arguments,
        # the signal and its group key
        else:
            for group, group_sigs in self._group_signals(signals).items():
                result = target(group_sigs, group, *args, **kwargs)
                if result:
                    output.extend(ensure_list(result))

        return output

    def process_signals(self, signals, input_id=None):
        """A convenient implementation of process signals for grouping blocks.

        This implementation of process_signals is provided only as a
        convenience to block developers. It prevents having to write the same
        code every time for process signals when in reality you only care about
        the signals after they are grouped.

        To use this, don't implement process_signals, but rather implement
        process_group_signals according to the signature provided in this file.
        Any signals you return from that method will be notified out of the
        block on the default output.

        If you wish to not use this implementation, just implement your own
        version of process_signals.
        """
        output_sigs = self.for_each_group(
            self.process_group_signals, signals, input_id)
        if output_sigs:
            self.notify_signals(output_sigs)

    def process_group_signals(self, signals, group, input_id=None):
        """Implement this method to make use of the default behavior.

        If you wish to make use of the default implementation of
        process_signals then override this method. If you wish, you may return
        a list of signals or an individual signal out of this method to have
        them notified out of the block.
        """
        raise NotImplementedError

    def _group_signals(self, signals):
        """ Groups the provided signals according to the configuration

        Args:
            signals (list): A list of signals to group

        Returns:
            dict: A dictionary where keys are group names and values are lists
                of the signals in that group
        """
        signal_groups = defaultdict(list)
        for s in signals:
            key = self.group_by(s)

            # Need to make sure that the key is a hashable object
            if not isinstance(key, Hashable):
                self.logger.warning(
                    "{} is not hashable cannot be grouped by. Using str "
                    "representation instead".format(key))
                key = str(key)

            # Add the key to our set of groups and the signal to the list
            # corresponding to the group
            self._groups.add(key)
            signal_groups[key].append(s)

        return signal_groups

    def groups(self):
        return self._groups

    def _groups_command(self):
        return {"groups": list(self._groups)}



from nio.modules.module import ModuleNotInitialized
from nio.modules.proxy import ProxyNotProxied
from nio.util.cache import Cache


class LogCache(object):

    """ A cache of log messages to prevent logging loops.

    Maintains logging instructions per code source line helping
    avoid potential infinite loops when logging from a logging handler
    subscriber.

    The criteria consists of determining if a log record from the same
    source code line and with the same message is in the Cache.
    """

    def __init__(self, expire_interval):
        """ Create a new LogCache instance.

        Args:
            expire_interval (seconds): how long an entry is kept in cache

        """
        self._cache = Cache(expire_interval)

    def process_record(self, record):
        """ Processes a record against cache

        A record gets added to the cache when a record from same source code
        line containing the same message does not already exists in the cache

        Args:
            record: log record as incoming from logging module

        Returns:
            True: if record is present in cache
        """

        # determine key, which includes a hashed msg, and attempt to get record
        key = "{}-{}-{}".format(record.filename, record.lineno,
                                hash(str(record.msg)))
        # determine if item with same message is present
        if self._cache.get(key):
            return True

        # record is always added or updated in cache under given key
        try:
            self._cache.add(key, record)
        except (ProxyNotProxied, ModuleNotInitialized):
            # catch any potential errors arising from the fact that
            # logging module starts before any other, i.e., scheduling
            pass

        return False



from nio.modules.security import User
from nio.modules.security.task import SecureTask
from nio.modules.security.authorizer import Unauthorized


class Authorizer(object):

    @classmethod
    def authorize(cls, user, task):
        if not isinstance(user, User) or not isinstance(task, SecureTask):
            raise Unauthorized()



from nio.testing.modules.scheduler.scheduler import JumpAheadScheduler


class JumpAheadJob(object):

    def __init__(self, target, delta, repeatable, *args, **kwargs):
        self._job = JumpAheadScheduler.schedule_task(
            target, delta, repeatable, *args, **kwargs)

    def cancel(self):
        JumpAheadScheduler.unschedule(self._job)

    def jump_ahead(self, seconds):
        """ Jump the scheudler forward a certain number of seconds.

        This is useful in tests to simulate time passing for event-driven
        logic and temporal assertions.
        """
        JumpAheadScheduler.jump_ahead(seconds)



import logging

from nio.util.logging.handlers.publisher.log_signal import LogSignal


class LogSignalFilter(logging.Filter):

    def filter(self, record):
        """  Filters out LogSignal records

        Args:
            record: record being filtered

        Returns:
            True if record should be logged, False otherwise.
        """
        return not isinstance(record.msg, LogSignal)



from nio.modules.web.server import WebServer
from nio.modules.web.engine import WebEngine
from nio.modules.web.handlers.web import WebHandler
from nio.modules.web.handlers.static import StaticHandler
from nio.modules.web.handlers.rest import RESTHandler



from collections import defaultdict
from datetime import timedelta, datetime
from threading import RLock

from nio.modules.scheduler.job import Job
from nio.signal.management import ManagementSignal
from nio.util.runner import Runner


class DiagnosticManager(Runner):

    _type_separator = "###"

    def __init__(self):
        super().__init__()

        self._start_time = None
        self._instance_id = None
        self._service_id = None
        self._service_name = None
        self._interval = None
        self._mgmt_signal_handler = None
        self._job = None

        self._blocks_data_lock = RLock()
        self._blocks_data = None

    def configure(self, context):
        self._instance_id = context.instance_id
        self._service_id = context.service_id
        self._service_name = context.service_name
        self._interval = \
            context.settings.get("diagnostic_interval", 3600)
        self._mgmt_signal_handler = context.mgmt_signal_handler
        self._blocks_data = defaultdict(lambda: defaultdict(int))

    def start(self):
        super().start()

        # initialize collection start time
        self._start_time = self._create_timestamp()
        # create job to periodically send updates
        if self._interval > 0:
            self._job = \
                Job(self._send_diagnostic,
                    timedelta(seconds=self._interval),
                    True)

    def stop(self):
        if self._job:
            self._job.cancel()

        # Send what we have saved
        self._send_diagnostic()
        super().stop()

    def on_signal_delivery(self,
                           source_type, source,
                           target_type, target, count):
        with self._blocks_data_lock:
            source_key = self._create_key(source_type, source)
            target_key = self._create_key(target_type, target)
            self._blocks_data[source_key][target_key] += count

    def _send_diagnostic(self):
        with self._blocks_data_lock:
            end_time = self._create_timestamp()
            if self._blocks_data and self._mgmt_signal_handler:
                blocks_data = []
                for source_key, target_data in self._blocks_data.items():
                    for target_key, count in target_data.items():
                        source_type, source = \
                            self._split_key(source_key)
                        target_type, target = \
                            self._split_key(target_key)
                        blocks_data.append(
                            {
                                "source_type": source_type,
                                "source": source,
                                "target_type": target_type,
                                "target": target,
                                "count": count
                            }
                        )
                self._mgmt_signal_handler(
                    ManagementSignal(
                        {
                            "type": "RouterDiagnostic",
                            "instance_id": self._instance_id,
                            "service_id": self._service_id,
                            "service": self._service_name,
                            "blocks_data": blocks_data,
                            "start_time": self._start_time,
                            "end_time": end_time
                        }
                    )
                )
                self._blocks_data.clear()
            self._start_time = end_time

    @staticmethod
    def _create_timestamp():
        """ Creates a calculated UTC timestamp.

        This is needed because datetime.utcnow().timestamp() does not
        play nice with timezones, so we calculate it manually.

        Returns:
            timestamp
        """
        return (datetime.utcnow() - datetime(1970,1,1)).total_seconds()

    @staticmethod
    def _create_key(_type, name):
        """ Creates a key by merging type and name

        Returns:
            created key
        """
        return "{}{}{}".format(_type, DiagnosticManager._type_separator, name)

    @staticmethod
    def _split_key(_id):
        """ Splits a key

        Returns:
            type, name tuple
        """
        ids = _id.split(DiagnosticManager._type_separator)
        return ids[0], ids[1]
